from pyqalx.core.adapters.adapter import QalxUnpackableAdapter
from pyqalx.core.entities import Group, Set


class QalxGroup(QalxUnpackableAdapter):
    _entity_class = Group
    child_entity_class = Set

    def add(self, sets, meta=None, **kwargs):
        """
        When adding a `Group` ensure that the sets posted to the api are in
        the format {<key>: pyqalx.entities.Set}

        :param sets: A dictionary of Sets to create on the group
        :type sets: dict
        :param meta: A dictionary of metadata to store
        :type meta: dict
        :return: A newly created `Group` instance
        """
        return super(QalxGroup, self).add(sets=sets, meta=meta, **kwargs)



from io import IOBase


try:
    from json.decoder import JSONDecodeError
except ImportError:
    # JSONDecodeError is for Python3 only
    JSONDecodeError = ValueError

import requests


class PyQalxAPIException(Exception):
    """
    A generic error for PyQalxAPI
    """


class BasePyQalxAPI(object):
    def __init__(self, config):
        self.config = config
        self.base_url = self.config.get("BASE_URL", "https://api.qalx.io/")
        self.token = self.config["TOKEN"]
        self._last_response = None

    def _build_request(self, url, method, include_auth_headers=True, **kwargs):
        """
        Handles making the request
        :param url: The url to make the request to
        :param include_auth_headers: Should we include the auth headers
        :param kwargs: kwargs to pass through to the request
        :return: response
        """
        if include_auth_headers:
            # We might not include the headers if we are PUTing to S3
            headers = kwargs.get("headers", {})
            # Don't overwrite existing headers
            headers.update(
                {"Authorization": "Token {token}".format(token=self.token)}
            )
            kwargs["headers"] = headers

        retry_count = 0
        MAX_RETRIES = self.config["MAX_RETRY_500"]
        if MAX_RETRIES < 1:
            raise PyQalxAPIException(
                f"`MAX_RETRY_500` config setting must "
                f"be at least `1`.  Got `{MAX_RETRIES}`"
            )
        while retry_count <= MAX_RETRIES:
            # There may be instances where the API returns a 500 error.  This
            # could be due to many things (network issues, cold starts etc).
            # To avoid terminating workers we retry up to `MAX_RETRIES` times
            # to make our best attempt at completing the request successfully.
            try:
                resp = requests.request(url=url, method=method, **kwargs)
                self._last_response = resp
                if resp.status_code < 500:
                    break
            except requests.exceptions.RequestException as exc:
                # RequestException: The base exception that `requests` can
                #                   raise. Catch everything and return to the
                #                   client
                resp = exc
            # We retry if the response was a 5XX or if there was a requests
            # exception of some sort
            retry_count += 1
        return self._build_response(resp)

    def _build_non_safe_request(self, method, endpoint, **kwargs):
        """
        Helper method for performing non safe requests (POST, PUT, PATCH etc).
        Handles the possibility of a user trying to create or update an item
        that also has a file
        :param method: The HTTP method
        :param endpoint: The endpoint you want to query: 'item`
        :param kwargs: Any data you want to pass to the request.
                      Use `json` key for data you want to post. Anything else
                      will be passed to `requests` as kwargs
                      json={'data': {'some': 'data'}, 'meta': {'some': 'meta'}}
        :return: response
        """
        url = self._get_url(endpoint)
        resp_ok, data = self._build_request(url=url, method=method, **kwargs)
        return resp_ok, data

    @staticmethod
    def _build_response(resp):
        """
        Builds the response that gets sent back to the client
        :param resp: The response from `requests` or an instance
                     of `RequestException`
        :return: tuple of `(response_ok:bool, response data:dict)`
        """
        # If it doesn't have `ok` then a `requests` exception was raised
        is_ok = getattr(resp, "ok", False)
        try:
            data = resp.json()
        except (AttributeError, JSONDecodeError):
            # AttributeError: `resp` is an instance of RequestException
            # JSONDecodeError: `resp` is a normal response but has no data.
            #                   i.e. a 500 or a `delete` response
            data = []
        if is_ok is False:
            # If either of these are missing then a `requests` exception was
            # raised
            status_code = getattr(resp, "status_code", "")
            reason = getattr(resp, "reason", resp)
            data = {
                "status_code": status_code,
                "reason": reason,
                "errors": data,
            }
        return is_ok, data

    def _get_url(self, endpoint):
        """
        Builds the URL for the request from the base_url and the endpoint
        :param endpoint: The endpoint you want to query: 'item/<item_guid>'
        :return: url
        """
        url = "{base_url}{endpoint}".format(
            base_url=self.base_url, endpoint=endpoint.rstrip("/").lstrip("/")
        )
        return url

    def _upload_to_s3(self, data, input_file):
        """
        Given a response object with a file url in will attempt to
        PUT the file onto S3

        :param data: The data from the api
        :param input_file: The path to the file to upload or the file itself
        :return: response
        """
        url = data["file"]["put_url"]
        if not self.is_filestream(input_file):
            file_data = open(input_file, "rb").read()
        else:
            file_data = input_file
        return self._build_request(
            url=url, method="PUT", include_auth_headers=False, data=file_data
        )

    @staticmethod
    def is_filestream(input_object):
        return isinstance(input_object, IOBase)



import os

import requests

from pyqalx.core.entities import QalxEntity, QalxListEntity
from pyqalx.core.errors import QalxError


class Item(QalxEntity):
    """An item is the core of qalx.

    They are structured data or a file or some combination of the two. For example:

    >>> from pyqalx import QalxSession
    >>> qalx = QalxSession()
    >>> dims = qalx.item.add(data={"height":5, "width":5}, meta={"shape":"square"})
    >>> steel = qalx.item.add(input_file="path/to/316_datasheet.pdf",
    ...                       data={"rho":8000, "E":193e9},
    ...                       meta={"library":"materials",
    ...                             "family":"steel",
    ...                             "grade":"316"})

    We can then use the ``find_one`` and ``find`` methods to search for items

    >>> from pyqalx import QalxSession
    >>> qalx = QalxSession()
    >>> steel_316_item = qalx.item.find_one(meta=["library=materials",
    ...                                           "family=steel",
    ...                                           "grade=316"])
    >>> steels = qalx.item.find(meta="family=steel")
    >>> squares = qalx.item.find(meta="shape=square")
    >>> quads = qalx.item.find(meta="(shape=square|shape=rectangle)")

    We can edit an item once we have retrieved it and save it back to qalx.
    You can either use attribute style getters/setters (my_shape.data.height = 10)
    or key style getters/setters (my_shape['data']['height'] = 10)

    >>> from pyqalx import QalxSession
    >>> qalx = QalxSession()
    >>> my_shape = qalx.item.find_one(data=["height=5", "width=5"])
    >>> my_shape.data.height = 10
    >>> my_shape.meta.shape = 'rectangle'
    >>> qalx.item.save(my_shape)
    >>> # If we think that someone else might have edited my_shape we can reload it:
    >>> my_shape = qalx.item.reload(my_shape)
    """

    entity_type = "item"
    _file_bytes = None

    def read_file(self):
        """
        If this Item contains a file, will read the file data and cache it
        against the Item.

        :return: The content of the URL as a bytes object.  Accessible from
            the `_file_bytes` attribute
        :raises: pyqalx.errors.QalxError
        """
        if not self.get("file"):
            raise QalxError("Item doesn't have file data.")
        else:
            response = requests.get(url=self["file"]["url"])
            if response.ok:
                self.__super_setattr__("_file_bytes", response.content)
                return self._file_bytes
            else:
                raise QalxError(
                    "Error with file retrieval: \n\n" + response.text
                )

    def save_file_to_disk(self, filepath, filename=None):
        """
        If this Item contains a file, will read the file from the URL (or from
        the cached bytes on the instance) and save the file to disk.  Provide
        an optional `filename` argument if you don't want to use the same
        filename as the one stored on the Item

        :param filepath: The path where this file should be saved
        :type filepath: str
        :param filename: The optional name of this file. Defaults to the name
            of the file on the instance
        :type filename: str
        :raises: pyqalx.errors.QalxError
        """
        if filename is None:
            filename = self["file"]["name"]
        if self._file_bytes is None:
            self.read_file()
        with open(os.path.join(filepath, filename), "wb") as f:
            f.write(self._file_bytes)


class ItemAddManyEntity(QalxListEntity):
    _data_key = "items"



from importlib import import_module

import click

from pyqalx import Bot, QalxSession
from pyqalx.config import UserConfig, BotConfig
from pyqalx.core.entities import QalxEntity


class QalxImport(click.ParamType):
    """
    A custom type to handle how the cli does imports.
    Expects a string in the format `my.dotted.module:attr_to_import`.
    """

    name = "import path"

    def __init__(self, expected_class, is_instance=False, *args, **kwargs):
        """
        :param expected_class: The class that this import should be an
        instance/subclass of
        :param is_instance: Should we do an `isinstance` check?  Defaults to
        doing an `issubclass` check
        """
        self.expected_class = expected_class
        self.is_instance = is_instance
        super(QalxImport, self).__init__(*args, **kwargs)

    def convert(self, value, param, ctx):
        module_name, variable_name = value.split(":")
        try:
            module = import_module(module_name)
        except ImportError:
            self.fail(f"Failed to import `{module_name}`")
        if variable_name not in dir(module):
            self.fail(f"`{variable_name}` not found in `{module_name}`.")
        _import_target = getattr(module, variable_name)

        method = issubclass
        message = "a subclass"

        if self.is_instance:
            method = isinstance
            message = "an instance"
        if not method(_import_target, self.expected_class):
            self.fail(
                f"`{variable_name}` isn't {message} "
                f"of `{self.expected_class.__name__}`."
            )
        return _import_target


BOT_IMPORT = QalxImport(Bot, is_instance=True)
QALX_SESSION_IMPORT = QalxImport(QalxSession)
USER_CONFIG_IMPORT = QalxImport(UserConfig)
BOT_CONFIG_IMPORT = QalxImport(BotConfig)
QALX_ENTITY_IMPORT = QalxImport(QalxEntity)



from pyqalx.core.adapters.adapter import QalxNamedEntityAdapter
from pyqalx.core.entities import Queue


class QalxQueue(QalxNamedEntityAdapter):
    _entity_class = Queue
    _bot_only_methods = ["get_messages"]

    @property
    def _queue_params(self):
        """
        The configurable parameters for a `Queue`
        :return:
        """
        return {
            "VisibilityTimeout": int(self.session.config["MSG_BLACKOUTSECONDS"])
        }

    def add(self, name, meta=None, **kwargs):
        """
        Queues are created with a name.  This name is stored in the metadata
        of the `Queue` instance

        :param name: The name we want to assign the Queue
        :type name: str
        :param meta: A dictionary of metadata to store
        :type meta: dict
        :param kwargs: Any other kwargs we are setting on the Queue
        :return: A newly created `Queue` instance
        """
        return super(QalxQueue, self).add(
            parameters=self._queue_params, name=name, meta=meta, **kwargs
        )

    def get_messages(self, queue):
        """
        Gets the messages on the `Queue` instance

        :return: A list of `QueueMessage` instances
        """
        config = self.session.config
        max_msgs = int(config["Q_MSGBATCHSIZE"])
        visibility = int(config["MSG_BLACKOUTSECONDS"])
        waittime = int(config["MSG_WAITTIMESECONDS"])

        message = queue.get_messages(max_msgs, visibility, waittime)
        return message

    def get_by_name(self, name):
        """a single queue by name

        :param name: name of queue
        :type name: str
        :return: pyqalx.core.entities.Queue
        :raises: pyqalx.errors.QalxReturnedMultipleError,
                 pyqalx.errors.QalxEntityNotFound
        """
        return super(QalxQueue, self).get_by_name(name)

    def get_or_create(self, name, meta=None, **kwargs):
        """
        Gets a Queue by the given name or creates it if it doesn't exist

        :param name:
        :type name: str
        :param meta: metadata about the queue
        :return: pyqalx.core.entities.Queue
        """
        return super(QalxQueue, self).get_or_create(name, meta=meta)



import inspect


class ObjectDict(dict):
    """
    Makes a dictionary behave like an object, with attribute-style access.
    """

    sentinel = object()  # Unique instance that cannot be passed as an argument

    def __new__(cls, arg=sentinel):
        """
        This recursive dict subclass converts a dict into an ObjectDict,
        which can use attribute access to retrieve and set keys.
        Floats, ints, strings, None and existing ObjectDicts are simply
        returned. Lists are transformed into lists of ObjectDicts.
        The __init__ method will only be called for ObjectDicts,
        since only then is the argument class is the same as the returned
        value. When a dict is passed in, object creation is delegated to
        the superclass, and its contents are initialised in __init__.
        """
        if arg is cls.sentinel:
            return super().__new__(cls)
        elif isinstance(arg, list):
            return list(ObjectDict(x) for x in arg)
        elif (
            isinstance(arg, dict) and not isinstance(arg, ObjectDict)
        ) or inspect.isgenerator(arg):
            return super().__new__(cls, arg)
        else:
            # Just return the object as it is if we can't figure out how to
            # turn it into an ObjectDict
            return arg

    def __init__(self, arg=sentinel):
        """
        This method is only called after __new__ returns an ObjectDict
        instance. The __init__ method actually initialises the values
        in the underlying dict.
        """
        if isinstance(arg, ObjectDict):
            return  # Assume existing ObjectDicts do not need re-initialising
        if arg is self.sentinel:  # Called without args
            arg = {}

        super(ObjectDict, self).__init__(arg)

        for (k, v) in self.items():
            self[k] = ObjectDict(v)

    def __getattr__(self, name: str):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        value = ObjectDict(value)
        super().__setitem__(name, value)

    def __dir__(self):
        return sorted(self.keys())



import importlib
import logging
import sys
import traceback
from logging.config import dictConfig
from logging.handlers import RotatingFileHandler

from pyqalx.core.errors import QalxError


class QalxRotatingFileHandler(RotatingFileHandler):
    """
    We always force some defaults for a qalx logger
    """

    def __init__(self, *args, **kwargs):
        # 10 MB
        kwargs["maxBytes"] = kwargs.get("maxBytes", 10 * 1024 * 1024)
        kwargs["backupCount"] = kwargs.get("backupCount", 10)
        kwargs["mode"] = kwargs.get("mode", "w")
        super(QalxRotatingFileHandler, self).__init__(*args, **kwargs)


class QalxLogFormatter(logging.Formatter):
    """
    Sometimes our log messages contain extra data.  Use this formatter to allow
    all handlers to use the same format while also handling the potential extra
    data
    """

    def __init__(self, style="%", *args, **kwargs):
        kwargs[
            "fmt"
        ] = "%(asctime)s %(name)-12s %(levelname)-8s %(process)d %(message)s"
        super(QalxLogFormatter, self).__init__(*args, **kwargs)
        self.style = style
        self.original_style = self._style

    def format(self, record):
        dict_record = record.__dict__
        self._style = self.original_style
        if (
            "entity_type" in dict_record.keys()
            and "entity_guid" in dict_record.keys()
        ):
            existing_format = self._style._fmt
            # Typically this is added if a user decorated function encounters
            # an unhandled exception
            existing_format += (
                " Entity Type: `%(entity_type)s`"
                " Entity GUID: `%(entity_guid)s`"
            )
            self._style = logging._STYLES[self.style][0](existing_format)
        return super(QalxLogFormatter, self).format(record)


def build_default_config(LEVEL, LOG_FILE_PATH):
    """
    Builds the default config for pyqalx.

    :param LEVEL: The level from the `LOGGING_LEVEL` variable in the
                  users config
    :param LOG_FILE_PATH: The path to the log file to be used when using
        `filehandler`
    :return: dict: The default config for pyqalx to use.
    """
    filehandler = {
        "class": "pyqalx.core.log.QalxRotatingFileHandler",
        "filename": LOG_FILE_PATH,
        "formatter": "default",
        "level": LEVEL,
    }
    DEFAULT_CONFIG = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {"default": {"()": "pyqalx.core.log.QalxLogFormatter"}},
        "handlers": {
            "nullhandler": {"class": "logging.NullHandler", "level": LEVEL},
            "filehandler": filehandler,
        },
        "loggers": {
            "pyqalx.core": {"handlers": ["filehandler"], "level": LEVEL},
            "pyqalx.bot": {"handlers": ["filehandler"], "level": LEVEL},
            "pyqalx.config": {"handlers": ["filehandler"], "level": LEVEL},
            "pyqalx.transport": {"handlers": ["filehandler"], "level": LEVEL},
            # Users can do `logging.getLogger('pyqalx.integration') in their
            # pyqalx implementation to keep all pyqalx logs together
            "pyqalx.integration": {"handlers": ["filehandler"], "level": LEVEL},
        },
    }
    return DEFAULT_CONFIG


def configure_logging(config):
    """
    Given a `Config` instance will attempt to get the configurable logging
    values from the profile - falling back to sensible defaults.
    It then configures the logger with our defaults and then overrides as
    necessary depending on if the user has specified a custom logging config
    :param config: An instance of `Config`
    """
    logging_config_func = dictConfig

    LOGGING_LEVEL = config.get("LOGGING_LEVEL").upper()  # noqa
    LOG_FILE_PATH = config.get("LOG_FILE_PATH")

    valid_levels = logging._levelToName.values()

    assert LOGGING_LEVEL in valid_levels, (
        "LOGGING_LEVEL must be one of "
        "{VALID_LEVELS}.  "
        "Value is {VALUE}".format(
            VALID_LEVELS=", ".join(valid_levels), VALUE=LOGGING_LEVEL
        )
    )

    LEVEL = getattr(logging, LOGGING_LEVEL.upper())

    logging_config_func(
        build_default_config(LEVEL=LEVEL, LOG_FILE_PATH=LOG_FILE_PATH)
    )


def get_logger(exc_type, exc_value, exc_traceback):
    """
    From the unhandled exception (i.e. QalxError()) we build a
    TracebackException which we then inspect to determine the correct logger
    to log the error to
    """
    tb_exc = traceback.TracebackException(exc_type, exc_value, exc_traceback)
    path = tb_exc.stack[-1].filename
    loggers = [
        logging.getLogger(name) for name in logging.root.manager.loggerDict
    ]
    # Assume anything that isn't internal is running as a bot from the
    # users side of things
    logger_name = "pyqalx.bot"
    for logger in loggers:
        if (
            logger.name.startswith("pyqalx")
            and logger.name != "pyqalx.integration"
        ):
            # special case `pyqalx.integration`.  It isn't a module but we
            # need it so users can log themselves.
            module = importlib.import_module(logger.name)
            if module.__file__ == path:
                logger_name = logger.name
    return logging.getLogger(logger_name)


def exception_hook(exc_type, exc_value, exc_traceback, extra=None, logger=None):
    """
    We want to log any unhandled exception that occurs in pyqalx.  This
    includes `Qalx*` errors (i.e. because a user has misconfigured something)
    and also if they have made an error in one of their decorated functions
    (i.e. @bot.process).  This catches all errors and logs them and also
    prints out the traceback
    """
    if issubclass(exc_type, QalxError):
        # Only handle `Qalx*` Errors.  We don't want to consume anything the
        # user has done wrong.  This will however handle issues with decorated
        # bot functions as these are caught in ~bot.Bot and handled as a
        # QalxError.  (i.e. if the user does `int('a')` in a bot function the
        # exception will be logged and raised here.)
        if issubclass(exc_type, KeyboardInterrupt):
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
            return
        if logger is None:
            logger = get_logger(exc_type, exc_value, exc_traceback)

        logger.error(
            msg="Qalx Exception:",
            exc_info=(exc_type, exc_value, exc_traceback),
            extra=extra,
        )

        traceback.print_exception(
            exc_type, exc_value, exc_traceback, chain=False
        )
    else:
        # This is an exception elsewhere in the users codebase outside of qalx
        # or any bot function. We don't handle it.  Just reraise it for
        # the user to fix
        raise exc_type(exc_value)


sys.excepthook = exception_hook



import itertools

import click
from dateutil.parser import parse
from tabulate import tabulate

from pyqalx.core.entities.bot import Bot
from pyqalx.core.signals import QalxSignal


class QalxCLITabulation:
    """
    A helper class for building a table of entities to display to the user
    """

    prompt_text = None

    def __init__(self, qalx_session, name, entity_class, tabulate_single=False):
        """
        :param qalx_session: An instance of QalxSession
        :param name: The name of the entity that we are looking for
        :param entity_class: The entity class that is being tabulated
        :param tabulate_single: Whether we should tabulate if a single
        result is found or just return the entity
        """

        self.qalx_session = qalx_session
        self.name = name
        self.tabulate_single = tabulate_single
        self.entity_type = entity_class.entity_type

        # Get the entities for the prompt
        self.entities_for_prompt = self.get_entities_for_prompt()

    def query_kwargs(self):
        """
        By default there are no additional kwargs passed through to the query
        """
        return {}

    def get_entities_for_prompt(self):
        """
        Queries the API and returns the entities for the prompt based on
        `self.build_query()` and `self.query_kwargs`
        """
        query = self.build_query()
        entities = getattr(self.qalx_session, self.entity_type).find(
            query=query, **self.query_kwargs()
        )
        return entities["data"]

    def build_query(self):
        """
        Builds the query that will be used to query the api
        :return: A dict of the query, by default just the name will be queried
        """
        return {"name": self.name}

    @property
    def entity_type_plural(self):
        return f"{self.entity_type}s"

    def build_no_entities_message(self):
        """
        The message to be displayed if no entities are found matching the query
        """
        return f"No {self.entity_type_plural} found called `{self.name}`"

    @staticmethod
    def build_row(index, entity):  # pragma: no cover
        """
        A specific row on a table
        :param index: The index of this row
        :param entity: The entity instance
        :return: A list of data for the entity for a single row
        """
        raise Exception("Must be implemented by subclass")

    @staticmethod
    def build_headers():  # pragma: no cover
        """
        The headers for this table
        :return: A list of strings for the table
        """
        raise Exception("Must be implemented by subclass")

    def tabulate(self):
        """
        For the given `self.entities_for_prompt` will build a table using
        tabulate and display this to the user via `click.echo`
        """
        table = []

        # Massage the remaining entities into a nice format for presentation
        for index, entity in enumerate(self.entities_for_prompt, start=1):
            table.append(self.build_row(index, entity))

        # We then display the entities to the user in a list,
        # and potentially prompt them for an index for the specific one they
        # wish to interact with
        headers = self.build_headers()
        # Display a table of entities
        click.echo(tabulate(table, headers))
        return table

    def get_entity_or_display(self):
        """
        Will either return the entity or display a table of entities
        depending on the number of entities found and the value
        of `self.tabulate_single`.
        If no entities are found always return None
        """
        entities_for_prompt = self.entities_for_prompt
        if len(entities_for_prompt) == 1:
            # Only a single entity found matching the name and with active
            # workers.
            # Just return it without showing the user the table of entities
            if self.tabulate_single:
                self.tabulate()
            else:
                return entities_for_prompt[0]
        elif not len(entities_for_prompt):
            # No entities found.  Nothing to do!
            click.echo(self.build_no_entities_message())
            return
        else:
            # Many entities found matching the query.
            # Create a table to allow the user to pick which they want
            index = 0
            table = self.tabulate()
            if self.prompt_text:
                # Wait until they choose an option that is valid.
                # Only display this prompt if there is prompt text.
                # If there isn't prompt text then we probably just want to
                # show a table of entities (i.e. it's an info command)
                while index == 0 or index > len(table):
                    index = click.prompt(
                        f"Please choose a {self.entity_type} index"
                        f" to {self.prompt_text}",
                        type=int,
                    )
                # Get the entity from the index (-
                # 1 because the index the user chooses starts at 1)
                return self.entities_for_prompt[index - 1]


class BotTabulation(QalxCLITabulation):
    def __init__(self, *args, **kwargs):
        super(BotTabulation, self).__init__(entity_class=Bot, *args, **kwargs)

    @staticmethod
    def build_headers():
        return [
            "Index",
            "Name",
            "Status",
            "Platform",
            "Node",
            "No. Workers",
            "Created On",
            "Created By",
        ]

    @staticmethod
    def build_row(index, entity):
        created_on = parse(entity["info"]["created"]["on"]).strftime(
            "%d/%m/%Y %H:%M:%S"
        )
        row = [
            index,
            entity["name"],
            entity["status"],
            entity["host"].get("platform"),
            entity["host"].get("node"),
            len(entity["workers"]),
            created_on,
            entity["info"]["created"]["by"]["email"],
        ]
        return row


class TerminateBotTabulation(BotTabulation):
    prompt_text = "terminate"

    def build_query(self):
        query = super(TerminateBotTabulation, self).build_query()
        query["workers"] = {"$exists": True, "$not": {"$size": 0}}
        return query

    def query_kwargs(self):
        """
        Only return the workers field when querying for terminated bots
        as another query is required to check the status of each
        """
        query_kwargs = super(TerminateBotTabulation, self).query_kwargs()
        query_kwargs["fields"] = ["workers"]
        return query_kwargs

    def build_no_entities_message(self):
        message = super(
            TerminateBotTabulation, self
        ).build_no_entities_message()
        return message + f" that have workers that are not already terminated"

    @staticmethod
    def _filterfalse_signal(_bot):
        """
        itertools.filterfalse is used to exclude bots if a specific signal
        IS FALSE.
        Returns the status of this bots terminate signal.  If this signal
        is True (i.e. the bot is terminated) then it WON'T get returned
        to the user
        """
        return QalxSignal(_bot).terminate

    def get_entities_for_prompt(self):
        """
        For the termination prompt, only show bots that exist with workers
        that aren't terminated.
        """
        bots = super(TerminateBotTabulation, self).get_entities_for_prompt()
        potential_bots = []
        # Because we don't unpack when doing a `find` we need to do
        # an extra reload query for each bot so we can get the signal data
        for bot in bots:
            potential_bots.append(self.qalx_session.bot.reload(bot))

        def _has_workers_with_specific_signal(_bot):
            # For each bot, we only return it if any of the
            # workers ARE NOT on a status of `self._filter_signal(x)`.
            # As a separate function to avoid nested lambdas
            return list(
                itertools.filterfalse(
                    lambda x: self._filterfalse_signal(x), _bot["workers"]
                )
            )

        # We then filter the bots to only be those that have any workers that
        # don't have a self._filterfalse signal.
        entities_for_prompt = list(
            filter(_has_workers_with_specific_signal, potential_bots)
        )
        return entities_for_prompt


class StopBotTabulation(TerminateBotTabulation):
    prompt_text = "stop"

    def build_no_entities_message(self):
        message = super(
            TerminateBotTabulation, self
        ).build_no_entities_message()
        return (
            message + f" that have workers that are not already stopped "
            f"or terminated"
        )

    @staticmethod
    def _filterfalse_signal(_bot):
        """
        itertools.filterfalse is used to exclude bots if a specific signal
        IS FALSE.
        :return: If the bots stop signal is True (i.e. the bot is stopped)
        then the bot WON'T get returned to the user OR if the bots
        terminate signal is True (i.e. the bot is terminated)
         then the bot WON'T get returned to the user.
         Therefore, this will only return non terminated bots that are also
         not stopped
        """
        signal = QalxSignal(_bot)
        return signal.terminate is True or signal.stop is True


class ResumeBotTabulation(TerminateBotTabulation):
    prompt_text = "resume"

    def build_no_entities_message(self):
        message = super(
            TerminateBotTabulation, self
        ).build_no_entities_message()
        return message + f" that have workers that can be resumed"

    @staticmethod
    def _filterfalse_signal(_bot):
        """
        itertools.filterfalse is used to exclude bots if a specific signal
        IS FALSE.
        :return: if the bots stop signal is False (i.e. the bot is not stopped)
        then the bot WON'T get returned to the user OR if the bots terminate
        signal is True (i.e. the bot is terminated) then the bot WON'T get
        returned to the user.
        Therefore, this will only return non terminated bots that are also
        stopped
        """
        signal = QalxSignal(_bot)
        return signal.terminate is True or signal.stop is False



import logging
import os
import sys
import threading
import warnings
from functools import partial
from multiprocessing import Process
from time import sleep
from typing import NamedTuple

# We may need below, we may not. What can I say? Dill waters run deep.
import deprecation
import dill

from pyqalx.config import BotConfig
from pyqalx.core.adapters import QalxQueue, QalxWorker
from pyqalx.core.session import QalxSession
from pyqalx.core.errors import (
    QalxBotInitialisationFailed,
    QalxStepFunctionNotDefined,
)
from pyqalx.core.errors import QalxError, QalxConfigError
from pyqalx.core.log import (
    exception_hook,
    QalxRotatingFileHandler,
    QalxLogFormatter,
)
from pyqalx.core.signals import QalxSignal

logger = logging.getLogger(__name__)

dill.settings["recurse"] = True


def do_nowt(*args, **kwargs):
    pass


def do_nowt_right(*args, **kwargs):
    return True


def cannot_be_undefined(*args, **kwargs):
    """
    Raises an exception if a step function cannot be left undefined.
    Handled here to avoid a user trying to call the specific step function
    directly.
    """
    raise QalxStepFunctionNotDefined(
        f"`{kwargs['step_name']}` function must be defined either "
        f"as a decorator (`@mybot.{kwargs['step_name']}`) or by "
        f"overriding the function entirely "
        f"(`mybot.{kwargs['step_name']}_function = <process_function>`)"
    )


class QalxJob:
    """a Job to process

    :param worker_process: The instance of the worker that is running this job
    :type worker_process: ~pyqalx.bot.Worker
    """

    __slots__ = (
        "worker_adapter",
        "entity_adapter",
        "entity",
        "session",
        "state",
        "options",
        "progress",
        "warnings",
        "errors",
        "worker_logger",
        "_current_step",
        "_process",
        "_worker_entity",
        "_bot_entity",
        "_queue_message",
        "_step_results",
    )

    def __init__(self, worker_process):
        self._process = worker_process
        self.worker_adapter = worker_process.worker_adapter
        self._bot_entity = worker_process.bot_entity
        self._worker_entity = worker_process.api_entity
        self.session = worker_process.session
        self.worker_logger = worker_process.logger

        self._current_step = "class init"
        self._step_results = []

    @property
    def log(self):
        """
        An instance of `worker_logger` to enable logging from step functions
        """
        return self.worker_logger

    @property
    def e(self):
        """shortcut to QalxJob().entity"""
        return self.entity

    @property
    def s(self):
        """shortcut to QalxJob().session"""
        return self.session

    def stop_processing(self):
        """
        Stop the worker from processing.  Will cause the worker to wait until
        it receives a `resume` signal
        """
        self.worker_adapter.stop(
            entity=self._worker_entity, bot_entity=self._bot_entity
        )

    def resume_processing(self):
        """
        Resume the worker.
        """
        self.worker_adapter.resume(
            entity=self._worker_entity, bot_entity=self._bot_entity
        )

    def terminate(self, warm=True, requeue_job=False):
        """
        The job has been told to terminate the worker

        :param warm: If True will completely finish running through the
                     processes for the current job.  If False will exit as soon
                     as the current process function is complete
        :type warm: bool
        :param requeue_job: Should the job be requeued if we are terminating
                            with `warm=False`
        :type requeue_job: bool
        """
        self.worker_adapter.terminate(
            entity=self._worker_entity,
            bot_entity=self._bot_entity,
            warm=warm,
            requeue_job=requeue_job,
        )

    def publish_status(self, status="", status_string="", status_dict=None):
        """update the status of the worker

        :param status: short message about the worker status
        :type status: str
        :param status_string: short message about the worker status
        :type status_string: str
        :param status_dict: structured status data
        :type status_dict: dict
        """
        if status_string:  # pragma: no cover
            the_warning = deprecation.DeprecatedWarning(
                function="status_string",
                deprecated_in="0.8",
                removed_in="1.0",
                details="`status_string` will be renamed to `status` in "
                "future versions to maintain consistency with Bots",
            )
            warnings.warn(
                the_warning, category=DeprecationWarning, stacklevel=2
            )
            # If the user supplies both arguments through the deprecation
            # period we won't know which to use
            assert not all(
                [status, status_string]
            ), "Ensure only `status` or `status_string` is specified"
            if status_string and not status:
                # Ensure that we use the status they passed through as a kwarg
                status = status_string

        if status_dict is not None:  # pragma: no cover
            the_warning = deprecation.DeprecatedWarning(
                function="status_dict",
                deprecated_in="0.8",
                removed_in="1.0",
                details="Data stored in `status_dict` is not saved by the API."
                "The argument will be removed in future versions.",
            )
            warnings.warn(
                the_warning, category=DeprecationWarning, stacklevel=2
            )
        self.worker_adapter.update_status(
            entity=self._worker_entity,
            bot_entity=self._bot_entity,
            status=status,
        )

    def add_step_result(self, success=True, result_data=None):
        """indicate success and add info about the current step

        :param success: was the step successful or not
        :type success: bool
        :param result_data: data from the result
        :type result_data: dict
        """
        self._step_results.append(
            StepResult(
                success=success,
                step_name=self._current_step,
                result_data=result_data,
            )
        )

    @property
    def last_step_result(self):
        """
        The last :class:`~pyqalx.bot.StepResult` object (or None if no
        previous steps) for this job
        """
        if self._step_results:
            return self._step_results[-1]
        else:
            return None

    def get_entity(self, entity_type, entity_guid):
        """
        get an entity

        :param entity_type: the type of entity e.g. item, set, group
        :type entity_type: str
        :param entity_guid: the guid of the entity
        :type entity_guid: str
        :return: a child of pyqalx.core.entities.QalxEntity
        """
        adapter = getattr(self.session, entity_type)
        self.entity_adapter = adapter
        self.entity = adapter.get(entity_guid, config=self.session.config)
        return self.entity

    def add_item_data(self, item_key, data, meta=None):
        """
        helper method to add data item to the set

        :param item_key: key on the set to add the item as
        :type item_key: str
        :param data: data to add
        :type data: dict
        :param meta: metadata
        :type meta: dict

        :return: The specific :class:`~pyqalx.core.entities.item.Item` instance
                 that has been added
        """
        if not (self.entity.entity_type == "set"):
            raise TypeError(
                "Method add_item_data only works on jobs with a set entity."
            )
        self.entity["items"][item_key] = self.session.item.add(
            data=data, meta=meta
        )
        self.save_entity()
        return self.entity["items"][item_key]

    def save_entity(self):
        """
        Updates the API representation of this entity with the data
        currently stored in the job.

        :return: a child of :class:`~pyqalx.core.entities.entity.QalxEntity` of
                 the correct type based on the type of `self.entity`
        """
        # Always ensure that the entity on the adapter is fully up to date
        # with the entity on the job.
        self.entity_adapter.save(entity=self.entity)
        return self.entity

    def reload_entity(self):
        """
        Reloads `self.entity` from the API

        :return: a child of :class:`~pyqalx.core.entities.entity.QalxEntity` of
                 the correct type based on the type of `self.entity`
        """
        # Always ensure that the entity on the adapter is fully up to date
        # with the entity on the job.
        self.entity = self.entity_adapter.reload(entity=self.entity)
        return self.entity

    def delete_message(self):
        """
        Helper method for deleting a message from a Queue
        """
        return self._worker_entity._delete_message()


class StepResult(NamedTuple):
    """
    An NamedTuple representing the result of a single step function.
    """

    # The name of the step function that this instance refers to
    step_name: str = ""
    # Was this step successful or not?
    success: bool = False
    # A dict of data for the result of the step
    result_data: dict = {}


class Bot:
    __slots__ = (
        "name",
        "queue_name",
        "initialisation_function",
        "begin_function",
        "preload_function",
        "onload_function",
        "preprocess_function",
        "process_function",
        "precompletion_function",
        "postprocess_function",
        "onwait_function",
        "ontermination_function",
        "bot_entity",
        "bot_adapter",
        "qalx_session_class",
        "user_config_class",
        "bot_config_class",
        "entity_classes",
    )

    def __init__(self, bot_name):
        """
        :param bot_name: The unique name for this bot
        :type bot_name: str
        """
        self.name = bot_name

        # bot lifecycle functions
        self.initialisation_function = do_nowt_right
        self.begin_function = do_nowt
        self.preload_function = do_nowt
        self.onload_function = do_nowt
        self.preprocess_function = do_nowt

        # Use `partial` so we can ensure the name of the function always goes
        # to the exception.
        # TODO:Figure out a nicer way to get the step_name in here
        self.process_function = partial(
            cannot_be_undefined, step_name="process"
        )
        self.precompletion_function = do_nowt
        self.postprocess_function = do_nowt
        self.onwait_function = do_nowt
        self.ontermination_function = do_nowt

    def _collect_steps(self):
        steps = {}

        for fn in self.__dir__():
            if fn.endswith("_function"):
                fn_attr = getattr(self, fn)
                if fn_attr not in [do_nowt, do_nowt_right]:
                    if getattr(fn_attr, "func", None) == cannot_be_undefined:
                        # This function must be defined, call it to raise
                        # an exception pre fork.  Only functions using
                        # `partial` will have the `func` attr
                        fn_attr()
                    steps[fn] = dill.dumps(fn_attr)  # serialise functions
        return steps

    def start(
        self,
        queue_name,
        processes=1,
        qalx_session_class=None,
        user_config_class=None,
        bot_config_class=None,
        entity_classes=None,
        user_profile_name="default",
        bot_profile_name="default",
        skip_ini=False,
        add_queue=True,
    ):
        """
        :param queue_name: The unique name for the queue
        :type queue_name: str
        :param qalx_session_class: The QalxSession class that this Bot and
        Workers should use
        :param user_config_class: The UserConfig class that will be used when
        creating the Bot
        :type user_config_class: `~configs.UserConfig` or subclass
        :param bot_config_class: The BotConfig class that will be used when
        using the bot session
        :type: str
        :param entity_classes: A list of `QalxEntity` subclasses that you want
        this Bot to use once it is started
        :type: list
        :param user_profile_name: The user profile name that will be used when
                                  creating the Bot
        :type user_profile_name: string
        :param bot_profile_name: The bot profile name that will be used when
                                 using the bot session
        :type bot_profile_name: string
        :param skip_ini: Should reading from the `ini` file be skipped when
                         instantiating the user and bot sessions
        :type skip_ini: bool
        :param processes: The number of processes this bot should spawn
        :type processes: int
        :param add_queue: Should the queue be created if it doesn't already
                          exist
        :type add_queue: bool
        """

        self.queue_name = queue_name

        if entity_classes is None:
            entity_classes = []

        if qalx_session_class is None:
            qalx_session_class = QalxSession

        user_session = qalx_session_class(
            profile_name=user_profile_name,
            config_class=user_config_class,
            skip_ini=skip_ini,
        )

        if bot_config_class is None:
            bot_config_class = BotConfig

        # This configures logging to use the config settings specified
        # on the bot
        bot_session = qalx_session_class(
            profile_name=bot_profile_name,
            config_class=bot_config_class,
            skip_ini=skip_ini,
        )
        if add_queue:
            # We use `get_or_create` to stop the bot erroring because
            # the user didn't know the queue already existed
            the_warning = deprecation.DeprecatedWarning(
                function="add_queue",
                deprecated_in="0.7",
                removed_in="1.0",
                details="Bots now use `get_or_create` when given a "
                "`queue_name` so the user no longer has the "
                "responsibility to create the queue before starting "
                "the bot",
            )
            warnings.warn(
                the_warning, category=DeprecationWarning, stacklevel=2
            )

            queue = user_session.queue.get_or_create(name=self.queue_name)
        else:
            # We create the queue if it doesn't already exist
            queue = user_session.queue.get_or_create(name=self.queue_name)
        # Add the queue_name - this makes it easier to find bots looking at
        # specific queues.
        bot_session.config.update({"queue_name": self.queue_name})
        self.bot_entity = user_session.bot.add(
            name=self.name, config=bot_session.config
        )

        self.bot_adapter = bot_session.bot

        self._register_entities(entity_classes)

        if self.initialisation_function:
            init_result = self.initialisation_function(user_session)
            logger.debug(f"Initialisation function returns {init_result}")
            if not init_result:
                raise QalxBotInitialisationFailed(
                    "The initialisation function returned a False-like value."
                )
        steps = self._collect_steps()

        # We need to update the config with the Bot Token
        self.bot_adapter.session._update_config(
            config={"TOKEN": self.bot_entity["info"]["credentials"]["token"]}
        )

        self.bot_entity = self.bot_adapter.replace_workers(
            bot_entity=self.bot_entity, workers=processes
        )

        workers = []
        # We use the same session and adapter for all workers

        worker_adapter = QalxWorker(session=self.bot_adapter.session)

        for worker in range(processes):
            workers.append(
                Worker(
                    steps,
                    queue,
                    self.bot_entity,
                    self.bot_entity["workers"][worker],
                    worker_adapter,
                    index=worker,
                )
            )
            logger.debug(
                f"Created worker {worker} to read from {self.queue_name}"
            )
        for p in workers:
            p.start()
        for p in workers:
            p.join()
        logger.debug("Workers joined.")

    def _register_entities(self, entity_classes):
        """
        Registers `self.entity_classes` for this Bots session
        """
        for entity_class in entity_classes:
            self.bot_adapter.session.register(entity_class)

    def initialisation(self, func):
        """decorator to to register the initialisation

        `initialisation` is executed with a QalxSession instance
        authenticated againt the `user_config_class` __init__ argument
        :param func: function to be executed
        """
        self.initialisation_function = func

    def begin(self, func):
        """decorator to to register the begin

        `begin` is executed with a QalxJob instance
        :param func: function to be executed
        """
        self.begin_function = func

    def preload(self, func):
        """decorator executed before entity is loaded from qalx.

        `preload` is executed with a QalxJob instance
        :param func: function to be executed
        """
        self.preload_function = func

    def onload(self, func):
        """decorator to register the onload

        `onload` is executed with a QalxJob instance
        :param func: function to be executed
        """
        self.onload_function = func

    def preprocess(self, func):
        """decorator to register preprocess function

        `preprocess` is executed with a QalxJob instance
        :param func: function to register
        """
        self.preprocess_function = func

    def process(self, func):
        """decorator to register process function

        `process` is executed with a QalxJob instance
        :param func: function to register
        """
        self.process_function = func

    def precompletion(self, func):
        """decorator to register precompletion function

        `precompletion` is executed with a QalxJob instance
        :param func: function to register
        """
        self.precompletion_function = func

    def postprocess(self, func):
        """decorator to register postprocess function

        `postprocess` is executed with a QalxJob instance
        :param func: function to register
        """
        self.postprocess_function = func

    def onwait(self, func):
        """decorator to register onwait function

        `onwait` is executed with a QalxJob instance
        :param func: function to register
        """
        self.onwait_function = func

    def ontermination(self, func):
        """decorator to register ontermination function

        `ontermination` is executed with a QalxJob instance
        :param func: function to register
        """
        self.ontermination_function = func


class Worker(Process):
    """

    exit codes (start at 72590):

        - 72590 base exit code
        - 72591 warm exit
        - 72592 cold exit
        - 72593 config triggered (e.g. KILL_AFTER)
    """

    def __init__(
        self,
        steps,
        queue,
        bot_entity,
        api_entity,
        worker_adapter,
        index,
        **kwargs,
    ):
        super(Worker, self).__init__(**kwargs)
        self._steps = steps
        self.queue = queue
        self.bot_entity = bot_entity
        self.api_entity = api_entity
        self.worker_adapter = worker_adapter
        self.session = self.worker_adapter.session
        self.queue_adapter = QalxQueue(session=self.session)
        self.stop = False
        self.config = self.session.config
        self.index = index
        self.worker_signal = None

    def configure_logging(self, index):
        """
        Configures the logger for the Worker. Each Worker logs to its own
        file. This prevents issues with multiple processes trying to write
        to the same file
        :param index: The index of the worker to use in the filename
        :return: An instance of `logging.Logger`
        """
        directory = self.config["WORKER_LOG_FILE_DIR"]
        level = self.config["LOGGING_LEVEL"]
        if os.path.isdir(directory) is False:
            raise QalxConfigError(
                f"WORKER_LOG_FILE_DIR config value "
                f"not a valid directory. Got `{directory}`"
            )

        worker_logger = logging.getLogger(f"pyqalx.worker.{index}")

        filename = os.path.join(
            directory, f'.worker_{self.bot_entity["name"]}.{index}.log'
        )
        handler = QalxRotatingFileHandler(filename=filename)

        handler.setFormatter(QalxLogFormatter())
        handler.setLevel(level)
        worker_logger.addHandler(handler)
        worker_logger.setLevel(level)
        return worker_logger

    def _reload_signal(self):
        """
        Helper method for reloading the signal for this worker
        """
        self.worker_signal = self.worker_adapter.get_signal(
            self.api_entity, bot_entity=self.bot_entity
        )

    def _check_worker_signal(self):
        """
        Reload the worker signal and either terminates the worker or stops it
        """
        self._reload_signal()

        if self.worker_signal.terminate_warm:
            self._do_exit(**QalxSignal._warm_exit_update())
        else:
            self.stop = self.worker_signal.stop

    def _do_exit(self, update, code=QalxSignal.BASE_EXIT_CODE):
        self.job._current_step = "OnTerminate"
        onterm = self.steps.get("ontermination_function")
        if onterm:
            onterm(self.job)
        self.job.publish_status(update)
        sys.exit(code)

    def _check_cold_exit(self, reload_signal=True):
        """
        Reloads the worker signal and checks whether we should do a cold
        termination
        """
        if reload_signal:
            self._reload_signal()

        if self.worker_signal.terminate_cold:
            # The worker has been asked to cold exit for some reason.
            # Check if the job needs to go back on the queue or if we can
            # delete it
            if self.job._queue_message:
                if self.worker_signal.requeue_job:
                    self.job._queue_message._stop_heartbeat()
                else:
                    self._delete_message()
            self._do_exit(**QalxSignal._cold_exit_update())

    def _check_warm_exit(self):
        """
        Reloads the worker signal and checks whether we should do a warm
        termination
        """
        self._reload_signal()

        if self.worker_signal.terminate_warm:
            # The worker has been asked to warm exit for some reason.
            # The job will have finished and the message should have been
            # deleted.
            self._do_exit(**QalxSignal._warm_exit_update())

    def _delete_message(self):
        """
        Helper function for stopping any heartbeat on the job and
        deleting the message
        """
        if getattr(self.job, "_queue_message", None):
            self.job._queue_message._stop_heartbeat()
            self.queue.delete_message(self.job._queue_message)

    def _wait_with_spreading(self, spread_factor):
        """
        We want to wait for a while before trying something again and wait
        longer and longer every time up to a maximum. This is to help limit the
        request count to qalx while staying responsive.

        :return: the next spreading factor to use
        """

        time_to_wait = int(self.config["Q_WAITTIMESECONDS"])
        max_spreading = int(self.config["Q_WAITTIMEMAXSPREADING"])
        self.job._queue_message = None
        self.job._current_step = "OnWait"
        if self.steps.get("onwait_function") is not None:
            self.steps["onwait_function"](self.job)
        total_wait_seconds = time_to_wait * spread_factor
        sleep(total_wait_seconds)
        self.logger.debug(f"wait for {total_wait_seconds}s")
        self._check_worker_signal()
        self._check_cold_exit(reload_signal=False)  # we reloaded in prev line
        next_spread = spread_factor + 1
        # never spread beyond a max factor
        return min(max_spreading, next_spread)

    def _continue_processing(self):
        if self.job.last_step_result:
            return (not self.stop) and self.job.last_step_result.success
        else:
            return not self.stop

    def _check_queue_credentials(self):
        """
        Checks the security token expiration timestamp on the queue and
        refreshes it if necessary
        """
        if self.queue._credentials_expiring_soon():
            self.logger.debug(
                f"refreshing security credentials for queue "
                f'`{self.queue["name"]}`'
            )
            self.queue = self.queue_adapter.reload(self.queue)

    def _run_functions(self):
        # Each worker has its own dedicated logger
        self.logger = self.configure_logging(index=self.index)
        self.job = QalxJob(self)
        _startup = True
        # avoid pickle errors
        self.steps = {}
        for f_name, f_dill in self._steps.items():  # de-serialise functions
            self.logger.debug(f"try to undill step {f_name}")
            try:
                self.steps[f_name] = dill.loads(f_dill)
            except TypeError:
                self.steps[f_name] = f_dill

        self.logger.debug(f"running")
        if self.steps.get("begin_function"):
            self.job._current_step = "Begin"

            self.steps.get("begin_function")(self.job)
            self._check_worker_signal()
            self._check_cold_exit(reload_signal=False)

        kill_after = self.config["KILL_AFTER"]
        spread_factor = 1
        no_message_attempts = 0

        while True:
            # Check if the credentials need refreshing. They may have expired
            # due to `_wait_with_spreading` or for another machine related
            # issue (i.e. machine hibernate)
            self._check_queue_credentials()
            if self.stop:
                spread_factor = self._wait_with_spreading(spread_factor)
                self.logger.debug(
                    f"Stop signal on Worker waited with factor {spread_factor}"
                )
            else:
                q_msgs = self.queue_adapter.get_messages(queue=self.queue)
                self.logger.debug(f"got {len(q_msgs)} messages")
                if q_msgs:
                    for q_msg in q_msgs:
                        if not _startup:
                            self.job = QalxJob(self)
                        _startup = False
                        self.job._queue_message = q_msg
                        self.logger.debug(f"process {q_msg.body}")
                        self._process_functions()
                        self._check_worker_signal()
                    no_message_attempts = 0  # reset the attempt counter
                    spread_factor = 1  # reset any wait spreading
                else:
                    spread_factor = self._wait_with_spreading(spread_factor)
                    no_message_attempts += 1

            if (kill_after is not None) and (
                no_message_attempts >= int(kill_after)
            ):
                self._do_exit("KILL_AFTER reached", 72593)

    def _process_functions(self):
        def process_as_far_as_possible(functions):
            for fn, step_name in functions:
                if self._continue_processing():
                    self.job._current_step = step_name
                    try:
                        fn(self.job)
                    except Exception:
                        # This needs to be a bare exception to handle
                        # something going wrong with the user defined functions
                        _, exc_value, exc_traceback = sys.exc_info()
                        # Send as a `QalxError` to our handler otherwise we
                        # risk consuming exceptions from the users application
                        exception_hook(
                            exc_type=QalxError,
                            exc_value=exc_value,
                            exc_traceback=exc_traceback,
                            extra={
                                "entity_guid": self.job.entity["guid"],
                                "entity_type": self.job.entity.entity_type,
                            },
                            logger=self.logger,
                        )

                    # check if we need to cold exit after every function
                    self._check_cold_exit()
                    if self.config.getboolean("SAVE_INTERMITTENT"):
                        self.job.save_entity()
                else:
                    break
            if not self.config.getboolean("SKIP_FINAL_SAVE"):
                self.job.save_entity()

            # We delete the message from the queue once we've processed it
            self._delete_message()
            return self.job

        # PreLoad
        if self.steps.get("preload_function"):
            self.job._current_step = "PreLoad"
            self.steps.get("preload_function")(self.job)
            self._check_cold_exit()
        if self._continue_processing():
            # Process steps
            steps = (
                "OnLoad",
                "PreProcess",
                "Process",
                "PreCompletion",
                "PostProcess",
            )
            process_functions = [
                (self.steps.get(f"{s.lower()}_function"), s)
                for s in steps
                if self.steps.get(f"{s.lower()}_function")
            ]

            entity_type = self.job._queue_message.body["entity_type"]
            entity_guid = self.job._queue_message.body["entity_guid"]
            self.job.get_entity(entity_type, entity_guid)

            t = threading.Thread(
                target=process_as_far_as_possible, args=(process_functions,)
            )
            t.start()
            while t.is_alive():
                # Start a heartbeat while we process this message.
                visibility = self.config["MSG_BLACKOUTSECONDS"]
                self.job._queue_message._do_heartbeat(
                    visibility=visibility, queue=self.queue
                )
                # Ensure the credentials are still valid after the heartbeat.
                # Especially important for long running heartbeats.
                self._check_queue_credentials()
        else:
            self._delete_message()

    def run(self):
        """
        Run our custom functions.  Catch any exception
        that may occur with it and pass that to our custom
        uncaught exception handler
        """
        try:
            self._run_functions()
        except Exception:
            # This needs to be a bare exception to handle something going
            # wrong with the Worker
            _, exc_value, exc_traceback = sys.exc_info()
            # Send as a `QalxError` to our handler otherwise we
            # risk consuming exceptions from the users application
            exception_hook(
                exc_type=QalxError,
                exc_value=exc_value,
                exc_traceback=exc_traceback,
                logger=self.logger,
            )



import logging

from pyqalx.core.adapters.adapter import (
    QalxUnpackableAdapter,
    QalxNamedEntityAdapter,
    QalxUpdateStatusAdapter,
)
from pyqalx.core.entities import Bot
from pyqalx.core.entities.worker import Worker

logger = logging.getLogger(__name__)


class QalxBot(
    QalxUnpackableAdapter, QalxNamedEntityAdapter, QalxUpdateStatusAdapter
):
    _entity_class = Bot
    child_entity_class = Worker
    _user_only_methods = ["add"]
    _bot_only_methods = ["replace_workers"]

    def _should_unpack(self):
        return True

    def _kids_for_lookup(self, entity):
        """
        The kids on a bot are just a list of guids
        """
        return entity[self.kids]

    def _unpacked_entities_to_valid_children(self, entity, unpacked_entities):
        """
        The unpacked entities are just a list on a Bot, therefore
        just return them

        :param entity: The Bot entity
        :param unpacked_entities: The unpacked Worker entities
        :return: The unpacked Worker entities
        """
        return unpacked_entities

    def _child_list_response(self, child_adapter, entity, **kwargs):
        """
        A bot needs to pass itself through to the child adapter in order
        to correctly build the list endpoint for Worker.  Don't
        filter the workers by anything - always return all of them for
        the given bot

        :param child_adapter: The child_adapter
        :param entity: The `QalxBot` entity
        :return: The child_adapter list response
        """
        return child_adapter.find(bot_entity=entity)

    def _pack_kids_for_request(self, kwargs):
        """
        Bots don't get added with workers - so just bypass the packing
        of the kids
        :param kwargs: kwargs sent to `add` method
        :return: kwargs
        """
        return kwargs

    def get_kids_keys_to_save(self, kids_to_save):
        """
        The kids on a Bot are a list, whereas normal packed entities are a dict
        Therefore, return the kids guids as a list
        :param kids_to_save: The kids on the Bot (either fully inflated or
        just the guids)
        :type kids_to_save: list
        :return:A list of guids of the kids on this Bot
        """
        return [
            str(k["guid"]) if isinstance(k, dict) else k for k in kids_to_save
        ]

    def cache_original_kids(self, kids):
        """
        The kids on a Bot are a list, so we need to cache them as a list
        :param kids: The kids on the bot
        :type kids: list
        :return: A list of the kids
        """
        return [k for k in kids]

    def add(self, name, config, meta=None, **kwargs):
        """
        Creates a `Bot` instance.

        :param name: The name that this bot will be given
        :type name: str
        :param config: The bots config
        :type config: dict
        :param meta: A dictionary of metadata to store
        :type meta: dict
        :return: The newly created `Bot` instance
        """
        return super(QalxBot, self).add(
            host=self.session._host_info,
            name=name,
            meta=meta,
            config=config,
            **kwargs,
        )

    def replace_workers(self, bot_entity, workers):
        """
        Completely replaces any Workers on the given bot.  Will return the
        replaced workers in an unpacked state

        :param bot_entity: The ~entities.bot.Bot entity that is being changed
        :param workers: The number of workers that this bot should have
        :return: A ~entities.bot.Bot instance with the updated workers
        """
        guid = bot_entity["guid"]
        detail_endpoint = self.detail_endpoint(guid=guid)
        endpoint = f"{detail_endpoint}/replace-workers"
        logger.debug(
            f"replace workers {self.entity_class.entity_type} with"
            f" guid {guid} with {endpoint}"
        )

        entity = self._process_api_request("patch", endpoint, workers=workers)
        entity = self.entity_class(entity)
        entity = self._attempt_unpack(entity)
        return entity

    def _signal_workers(self, entity, signal_method, *args, **kwargs):
        """
        Helper method for calling signals on workers.  A bot doesn't have
        the concept of `signals`.  Instead, we call a signal method we just
        apply this to all the workers on the bot
        """

        for worker in entity.get(self.kids, []):
            worker_method = getattr(self.session.worker, signal_method)
            worker_method(worker, bot_entity=entity, *args, **kwargs)

    def stop(self, entity, **kwargs):
        """
        Stops all the workers on the bot
        """
        self._signal_workers(entity, signal_method="stop", **kwargs)

    def resume(self, entity, **kwargs):
        """
        Resumes all the workers on the bot
        """
        self._signal_workers(entity, signal_method="resume", **kwargs)

    def terminate(self, entity, *args, **kwargs):
        """
        Terminates all the workers on the bot
        """
        self._signal_workers(
            entity, signal_method="terminate", *args, **kwargs
        )  # noqa



import configparser
import os
from inspect import ismodule
from os import environ

from . import defaults
from pyqalx.config.defaults import bots
from pyqalx.core.errors import (
    QalxConfigProfileNotFound,
    QalxConfigFileNotFound,
    QalxConfigError,
)


class Config(dict):
    key_prefix = ""
    filename = ""
    BOOLEAN_STATES = {
        "1": True,
        "yes": True,
        "true": True,
        "on": True,
        "0": False,
        "no": False,
        "false": False,
        "off": False,
        True: True,
        False: False,
    }

    def __init__(self):
        dict.__init__(self, self.defaults)
        for key, value in [
            (k, v) for k, v in environ.items() if k.startswith(self.key_prefix)
        ]:
            self[key.replace(self.key_prefix, "")] = value

    @classmethod
    def config_path(cls):
        return os.path.join(os.path.expanduser("~"), cls.filename)

    def from_inifile(self, profile_name="default"):

        if os.path.exists(self.config_path()):
            config = configparser.ConfigParser()
            config.optionxform = str
            with open(self.config_path()) as cfg:
                config.read_string(
                    cfg.read()
                )  # this makes mocks easier over config.read(filepath)
            if profile_name in config.keys():
                config_dict = config[profile_name]
                self.update(config_dict)
            else:
                profiles = "\n".join(config.keys())
                msg = (
                    "Couldn't find profile named {} in {} file."
                    " Did you mean one of:\n{}".format(
                        profile_name, self.filename, profiles
                    )
                )
                raise QalxConfigProfileNotFound(msg)
        else:
            raise QalxConfigFileNotFound(
                "Couldn't find {}".format(self.config_path())
            )

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, dict.__repr__(self))

    def getboolean(self, key):
        """
        Helper method for allowing us to handle multiple boolean values
        """
        # Convert to a string in case we're being given a python bool()
        value = str(self[key])
        if value.lower() not in self.BOOLEAN_STATES:
            raise ValueError("Not a boolean: %s" % value)
        return self.BOOLEAN_STATES[value.lower()]

    @property
    def defaults(self):
        return {
            k: v
            for k, v in vars(defaults).items()
            if (not k.startswith("__")) and (not ismodule(v))
        }

    @classmethod
    def configure(cls, profile_name, config_items=None):
        """
        When given a profile name and dict of config items will write the
        config file to disk.  If a `profile_name` is given that already exists
        on the profile then the `profile_name` on the profile will be
        completely replaced with the values from config_items
        :param profile_name: The name of the profile to write
        :param config_items: A dict of items that should be written to the
        config
        """
        if config_items is None:
            # We might just be writing a blank config (i.e. for `.bots`)
            config_items = {}
        with open(cls.config_path(), "a") as cfg:
            config_string = f"[{profile_name}]\n"
            for config_key, config_value in config_items.items():
                config_string += f"{config_key}={config_value}\n"
            cfg.write(config_string)


class BotConfig(Config):
    """Works exactly like a dict but provides ways to fill it from ini files
    and environment variables.  There are two common patterns to populate the
    config.
    Either you can add them to the `.bots` file:
        bot.config.from_botsfile(profile_name="default")
    Or alternatively you can define the configuration from environment variables
    starting with `QALX_BOT_`. These will be populated automatically
    but values defined in `.bots` will overwrite these if
    `bot.config.from_botsfile()` is called.
    To set environment variables before launching the application you have to set this
    environment variable to the name and value you want to use.  On Linux and OS X
    use the export statement::
        export QALX_BOT_LICENCE_FILE='/path/to/licence/file'
    On windows use `set` instead.
    :param defaults: an optional dictionary of default values
    """

    key_prefix = "QALX_BOT_"
    filename = ".bots"

    @property
    def defaults(self):
        config = super(BotConfig, self).defaults
        bot_config = {
            k: v
            for k, v in vars(bots).items()
            if not k.startswith("__") and (not ismodule(v))
        }
        config.update(bot_config)
        return config

    def from_inifile(self, profile_name="default"):
        super(BotConfig, self).from_inifile(profile_name)
        # check if the the specified log directory exists. If not, try to create
        # it, and raise a QalxConfigError in the case of OSError or PermissionError
        if not os.path.isdir(self["WORKER_LOG_FILE_DIR"]):
            try:
                os.mkdir(self["WORKER_LOG_FILE_DIR"])
            except (OSError, PermissionError):
                msg = (
                    "The specified worker log directory: {}"
                    " does not exist and it could not be created."
                ).format(self["WORKER_LOG_FILE_DIR"])
                raise QalxConfigError(msg)


class UserConfig(Config):
    """Works exactly like a dict but provides ways to fill it from ini files
    and environment variables.  There are two common patterns to populate the
    config.
    Either you can add them to the `.qalx_bot` file:
        qalx_bot.config.from_qalxfile(profile_name="default")
    Or alternatively you can define the configuration from environment variables
    starting with `QALX_USER_`. These will be populated automatically
    but values defined in `.bots` will overwrite these if
    `qalx_bot.config.from_qalxfile()` is called.
    To set environment variables before launching the application you have to set this
    environment variable to the name and value you want to use.  On Linux and OS X
    use the export statement::
        export QALX_USER_EMPLOYEE_NUMBER=1280937
    On windows use `set` instead.
    :param defaults: an optional dictionary of default values
    """

    key_prefix = "QALX_USER_"
    filename = ".qalx"



import logging
import os


TOKEN = None
"""API token"""

LOGGING_LEVEL = logging._levelToName[logging.ERROR]
"""Level that we should log at by default. """

LOG_FILE_PATH = os.path.join(os.path.expanduser("~"), ".qalx.log")
"""The path where log files will be stored.

 .. note::
    The default value is evaluated on our build system.
    Your default location will be ~/.qalx.log or %USERPROFILE%/.qalx.log"""

UNPACK_SET = True
"""
Should the items in a set be de-referenced when doing a `get` or `find_one` query?

.. note::
    Sets are not unpacked when doing a `find` query
"""

UNPACK_BLUEPRINT = True
"""Should the item blueprints in a set blueprint be de-referenced?"""

UNPACK_GROUP = False
"""
Should the sets in a group be de-referenced when doing a `get` or `find_one` query?

.. note::
    Groups are not unpacked when doing a `find` query
"""

MSG_BLACKOUTSECONDS = 30
"""
After this time in seconds the message will be returned to the queue if
it has not been deleted.

.. note::
    Workers automatically issue a heartbeat to the queue while the job is being
    processed to keep extending the blackout during processing
"""

MAX_RETRY_500 = 2
"""How many times should pyqalx retry the request when receiving a 5XX error
back from the API?"""



from itertools import zip_longest

from pyqalx.core.entities.object_dict import ObjectDict

from pyqalx.core.errors import QalxAPIResponseError, QalxEntityTypeNotFound


class QalxListEntity(ObjectDict):
    """
    Simple wrapper around a pyqalxapi_dict so we can keep extra keys
    on the API list response.  Instantiates each entity in `data` to the
    correct QalxEntity subclass.
    """

    _data_key = "data"

    def __new__(cls, pyqalxapi_list_response_dict, *args, **kwargs):
        """
        A QalxListEntity is just an ObjectDict that has a list of
        QalxEntity instances stored on the `_data_key`.
        :param pyqalxapi_list_response_dict: A dict that gets returned from
        a pyqalxapi list endpoint.  This should at minimum have a `data`
        key but may have other keys which we preserve
        :param kwargs: Must contain `child` key which is a subclass of `QalxEntity`

        """
        cls.child = kwargs["child"]

        if not issubclass(cls.child, (QalxEntity,)):
            raise QalxEntityTypeNotFound(
                f"Expected `child` to be a subclass of "
                f"`QalxEntity`.  Got `{cls.child}`"
            )

        return super(QalxListEntity, cls).__new__(
            cls, pyqalxapi_list_response_dict
        )

    def __init__(self, pyqalxapi_list_response_dict, *args, **kwargs):
        super().__init__(pyqalxapi_list_response_dict)

        if (
            self._data_key not in pyqalxapi_list_response_dict
            or not isinstance(
                pyqalxapi_list_response_dict[self._data_key], list
            )
        ):
            raise QalxAPIResponseError(
                "Expected `{0}` key in "
                "`pyqalxapi_list_response_dict` and for"
                " it to be a list".format(self._data_key)
            )
        # Cast all the entities in data to be an instance of `self.child`
        self[self._data_key] = [
            self.child(e) for e in pyqalxapi_list_response_dict[self._data_key]
        ]  # noqa

    def __str__(self):
        return f"[{self.child.entity_type} list]"


class QalxEntity(ObjectDict):
    """Base class for qalx entities_response.

    QalxEntity children need to be populated with either a
    `requests.models.Response` which is the type returned by the methods
    on `pyqalxapi.api.PyQalxAPI` or with a `dict`.

    Entities can behave either like a dict or attribute lookups can be used
    as getters/setters

    >>> class AnEntity(QalxEntity):
    ...     pass
    >>> c = AnEntity({"guid":"123456789", "info":{"some":"info"}})
    >>> # dict style lookups
    >>> c['guid']
    '123456789'
    >>> # attribute style lookups
    >>> c.guid
    '123456789'


    :param pyqalxapi_dict: a 'dict' representing a qalx entity object to
        populate the entity
    :type pyqalxapi_dict: dict
    """

    entity_type: str

    def __init__(self, pyqalxapi_dict):
        super().__init__(pyqalxapi_dict)

    def __str__(self):
        return f"[{self.entity_type}] {self['guid']}"

    @classmethod
    def _chunks(cls, _iterable, chunk_size, fillvalue=None):
        """
        Collect data into fixed-length chunks or blocks"
        # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
        Taken from the itertools documentation
        """
        args = [iter(_iterable)] * chunk_size
        return zip_longest(fillvalue=fillvalue, *args)

    def __super_setattr__(self, name, value):
        """
        Convenience method for setting proper attributes on the entity class.
        Because we are using `ObjectDict` if we did
        `self.<name> = value` this would set the
        dict key of `<name>` which we don't want.  So we call
        the supermethod to properly set the attribute
        :param name: The name of the attribute to set
        :param value: The value of the attribute to set
        """
        super(ObjectDict, self).__setattr__(name, value)

    def __dir__(self):
        """
        By default `ObjectDict` __dir__ only returns the keys on the dict.
        We want it to return everything as normal as entities might have
        methods that the user needs to know about
        """
        return super(ObjectDict, self).__dir__()



import deprecation
import logging

from pyqalx.__version__ import __version__
from pyqalx.core.adapters.adapter import (
    QalxValidateBlueprintAdapter,
    QalxFileAdapter,
)
from pyqalx.core.entities import Item
from pyqalx.core.entities.item import ItemAddManyEntity
from pyqalx.core.errors import QalxFileError, QalxError


logger = logging.getLogger(__name__)


class QalxItem(QalxValidateBlueprintAdapter, QalxFileAdapter):
    _entity_class = Item

    def _validate(
        self,
        data=None,
        input_file=None,
        file_name="",
        meta=None,
        blueprint_name=None,
        **kwargs,
    ):
        if data is None:
            data = {}
        return super(QalxItem, self)._validate(
            data=data,
            input_file=input_file,
            file_name=file_name,
            meta=meta,
            blueprint_name=blueprint_name,
            **kwargs,
        )

    def add(
        self,
        data=None,
        input_file=None,
        file_name="",
        meta=None,
        blueprint_name=None,
        **kwargs,
    ):
        """
        Adds an `Item` instance that can contain either
        data (as a dict), a file or both.

        :param data: Optional data to store against this Item
        :type data: dict
        :param input_file: file path or instance of
            StringIO or BytesIO (https://docs.python.org/3/library/io.html)
        :type input_file: str or `io` object
        :param file_name: input file name. Optional if a file path is given
         for input_file.  Required if an `io` object
        :type file_name: str
        :param meta: A dictionary of metadata to store
        :type meta: dict
        :param blueprint_name: An optional blueprint name to use to validate
         this item against an existing Blueprint
        :return: A newly created `Item` instance
        """

        return super(QalxItem, self).add(
            data=data,
            input_file=input_file,
            file_name=file_name,
            meta=meta,
            blueprint_name=blueprint_name,
            **kwargs,
        )

    def save(
        self,
        entity,
        input_file=None,
        file_name="",
        blueprint_name=None,
        **kwargs,
    ):
        """
        Saves an updated existing `Item` instance.
        To remove a file from an `Item` update the entity so that
        `entity['file'] = {}`

        :param entity: The entity that we are saving
        :type entity: An instance of `Item`
        :param input_file: file path or instance of StringIO or
            BytesIO (https://docs.python.org/3/library/io.html)
        :type input_file: str or `io` object
        :param file_name: input file name. Optional if a file path is given
            for input_file.  Required if an `io` object
        :type file_name: str
        :type file_name: str
        :param meta: A dictionary of metadata to store
        :type meta: dict
        :param blueprint_name: An optional blueprint name to use if you want
                               to validate this item against an existing
                               Blueprint
        :type blueprint_name: str
        :return: An updated `Item` instance
        """
        return super(QalxItem, self).save(
            entity=entity,
            input_file=input_file,
            file_name=file_name,
            blueprint_name=blueprint_name,
            **kwargs,
        )

    def add_many(self, items, **kwargs):
        """
        Adds multiple items at once.
        :param items: A list of item data that you want to create.  This can
        be a mixture of non file items and file items. The keys of each item in
        the list should be the same as if a single item were being created via
        the `add` method
        :type items: list
        :return:
        """
        if not isinstance(items, list):
            raise QalxError("`items` argument must be a list")
        headers = {"qalx-bulk": "true"}
        validated_items = []

        for item in items:
            # Ensure all the items are valid.  They should be in the same
            # format as if a user is adding a single item.  This also handles
            # checking the blueprint for each item
            validated_items.append(self._validate(**item))

        resp = self.add(
            items=validated_items,
            headers=headers,
            _is_validated=True,
            # Don't upload as we need to handle
            # the file upload manually
            upload=False,
            **kwargs,
        )

        # We have all the data to create the files, but the indexes have
        # been split up compared to the response.  Therefore, inject blank
        # values into the response where there are errors.  This means that
        # the structure will match the original `items` that the user specified
        # (as the API preserves order when bulk creating data)
        ERRORED_KEY = None
        for _ in resp["errors"].keys():
            resp["items"].insert(int(_), ERRORED_KEY)

        for index, item in enumerate(items):
            # Iterate the original `items` provided by the user
            created_item = resp["items"][index]
            if item.get("input_file") and created_item is not None:
                # There was a file on the original item and it was also
                # successfully created by the API.  Upload the file as normal
                self._upload_file(
                    entity=resp["items"][index], input_file=item["input_file"]
                )
        # Finally, remove all the `None` values from the items as these were
        # errors that we are done with
        resp["items"] = list(
            filter(lambda x: x is not ERRORED_KEY, resp["items"])
        )
        if resp["errors"]:
            for index, error in resp["errors"].items():
                logger.warning(
                    f"Error bulk creating at index `{index}`. "
                    f"Error: `{error}`. "
                    f"Original Data: `{items[int(index)]}`"
                )
        return ItemAddManyEntity(resp, child=self.entity_class)

    @deprecation.deprecated(
        deprecated_in="0.6",
        removed_in="1.0",
        current_version=__version__,
        details="Use the `add` function instead",
    )
    def add_file(
        self,
        input_file,
        data=None,
        meta=None,
        file_name="",
        blueprint_name=None,
        **kwargs,
    ):
        """adds a file to qalx with optional data and metadata

        :param input_file: input file path or stream
        :type input_file: str or `filestream`
        :param data: optional data you want to store with this file
        :type data: dict
        :param meta: additional data about the item
        :param meta: dict
        :param file_name: input file name. Optional for file path. Required
                          for file stream
        :type file_name: str
        :param blueprint_name: An optional blueprint name to use if you want
                              to validate this item against an existing
                              Blueprint
        :type blueprint_name: str
        :return: pyqalx.core.Item
        """
        if data is None:
            data = {}

        if self.session.rest_api.is_filestream(input_file) and not file_name:
            raise QalxFileError(
                "You must supply a file name when supplying a file stream"
            )
        return self.add(
            input_file=input_file,
            file_name=file_name,
            data=data,
            meta=meta,
            blueprint_name=blueprint_name,
            **kwargs,
        )



from os import getcwd

import click
import sys
from click import ClickException, types

from pyqalx import QalxSession
from pyqalx.config import UserConfig, BotConfig
from pyqalx.core.errors import QalxConfigProfileNotFound, QalxConfigFileNotFound
from qalxcli import cli_types
from qalxcli.utils import (
    TerminateBotTabulation,
    StopBotTabulation,
    ResumeBotTabulation,
    BotTabulation,
)


@click.group(chain=True)
@click.option(
    "-u",
    "--user-profile",
    help="User profile name in .qalx.",
    default="default",
)
@click.option(
    "-b", "--bot-profile", help="Bot profile name in .bots.", default="default"
)
@click.version_option()
@click.pass_context
def qalx(ctx, user_profile, bot_profile):
    """Command line interface to qalx."""
    ctx.ensure_object(dict)
    ctx.obj["USER_PROFILE"] = user_profile
    ctx.obj["BOT_PROFILE"] = bot_profile
    sys.path.append(getcwd())


@qalx.command("start-bot")
@click.option(
    "-p",
    "--processes",
    help="The number of workers to spawn in the bot.",
    default=1,
)
@click.option(
    "--skip-ini/--no-skip-ini",
    help="Should the bot use the profiles on disk?",
    default=False,
)
@click.option(
    "--add-queue/--no-add-queue",
    help="Should the bot try and create the queue if it doesn't exist?",
    default=True,
)
@click.argument("target", type=cli_types.BOT_IMPORT)
@click.option(
    "-q",
    "--queue-name",
    help="The name of the queue this bot should read messages from",
    required=True,
)
@click.option(
    "--qalx-session-class",
    type=cli_types.QALX_SESSION_IMPORT,
    help="The import path of a custom QalxSession class in the format "
    "`dotted.path.to:MyQalxSessionClass`",
    default=None,
)
@click.option(
    "--user-config-class",
    type=cli_types.USER_CONFIG_IMPORT,
    help="The import path of a custom UserConfigClass class in the format "
    "`dotted.path.to:MyUserConfigClass`",
    default=None,
)
@click.option(
    "--bot-config-class",
    type=cli_types.BOT_CONFIG_IMPORT,
    help="The import path of a custom BotConfigClass class in the format "
    "`dotted.path.to:MyBotConfigClass`",
    default=None,
)
@click.option(
    "--entity-class",
    type=cli_types.QALX_ENTITY_IMPORT,
    help="The import path of a custom QalxEntity class in the format "
    "`dotted.path.to:MyQalxEntityClass`",
    multiple=True,
)
@click.pass_context
def start(
    ctx,
    target,
    queue_name,
    processes,
    skip_ini,
    add_queue,
    qalx_session_class,
    user_config_class,
    bot_config_class,
    entity_class,
):
    """
    start the bot at TARGET

    This is the import path with a colon followed by the variable name
    of the bot. e.g. `my_qalx.my_bot:bot`
    """
    click.echo(
        f"Starting {target.name} reading from {queue_name} with"
        f" {processes} workers."
    )
    user_profile = ctx.obj.get("USER_PROFILE", "default")
    bot_profile = ctx.obj.get("BOT_PROFILE", "default")
    target.start(
        queue_name=queue_name,
        user_profile_name=user_profile,
        bot_profile_name=bot_profile,
        processes=int(processes),
        skip_ini=skip_ini,
        add_queue=add_queue,
        qalx_session_class=qalx_session_class,
        user_config_class=user_config_class,
        bot_config_class=bot_config_class,
        entity_classes=list(entity_class),
    )


@qalx.command("terminate-bot")
@click.argument("name")
@click.pass_context
def terminate_bot(ctx, name):
    """shutdown the bot named NAME"""
    qalx_session = QalxSession(
        profile_name=ctx.obj.get("USER_PROFILE", "default")
    )
    bot_tabulation = TerminateBotTabulation(qalx_session, name)
    bot_to_kill = bot_tabulation.get_entity_or_display()

    if bot_to_kill:
        # Finally, terminate the bot
        qalx_session.bot.terminate(bot_to_kill)
        click.echo(
            f"Terminate signal sent for `{bot_to_kill['name']} "
            f"({bot_to_kill['host'].get('platform')})`.  Workers may take "
            f"some time to terminate"
        )


@qalx.command("stop-bot")
@click.argument("name")
@click.pass_context
def stop_bot(ctx, name):
    """stop the bot named NAME.

    All workers to stop pulling jobs from the queue.
    """
    qalx_session = QalxSession(
        profile_name=ctx.obj.get("USER_PROFILE", "default")
    )
    bot_tabulation = StopBotTabulation(qalx_session, name)
    bot_to_stop = bot_tabulation.get_entity_or_display()

    if bot_to_stop:
        qalx_session.bot.stop(bot_to_stop)
        click.echo(f"Stopping {name}.")


@qalx.command("resume-bot")
@click.argument("name")
@click.pass_context
def resume_bot(ctx, name):
    """resume the bot named NAME.

    All workers to start pulling jobs from the queue.
    """
    qalx_session = QalxSession(
        profile_name=ctx.obj.get("USER_PROFILE", "default")
    )
    bot_tabulation = ResumeBotTabulation(qalx_session, name)
    bot_to_resume = bot_tabulation.get_entity_or_display()
    if bot_to_resume:
        qalx_session.bot.resume(bot_to_resume)
        click.echo(f"Resuming {name}.")


@qalx.command("bot-info")
@click.argument("name")
@click.pass_context
def bot_info(ctx, name):
    """print info about the bot named NAME.

    """
    qalx_session = QalxSession(
        profile_name=ctx.obj.get("USER_PROFILE", "default")
    )
    bot_tabulation = BotTabulation(qalx_session, name, tabulate_single=True)
    bot_tabulation.get_entity_or_display()


@qalx.command("terminate-workers")
@click.argument("bot_name")
@click.argument("number", type=types.INT)
@click.pass_context
def terminate_workers(ctx, bot_name, number):
    """terminate NUMBER of workers on bot named BOT_NAME."""
    qalx_session = QalxSession(
        profile_name=ctx.obj.get("USER_PROFILE", "default")
    )
    bot_tabulation = TerminateBotTabulation(qalx_session, bot_name)
    bot_to_kill = bot_tabulation.get_entity_or_display()
    if bot_to_kill:
        if number > len(bot_to_kill["workers"]):
            raise ClickException(
                f"{number} is more than the number of workers on {bot_name}"
                f" ({len(bot_to_kill['workers'])})"
            )
        for n in range(number):
            qalx_session.worker.terminate(
                bot_to_kill["workers"][n], bot_entity=bot_to_kill
            )
            click.echo(f"Terminated worker number {n + 1}.")


@qalx.command("configure")
@click.option("--user/--no-user", default=True)
@click.option("--bot/--no-bot", default=True)
@click.argument("extra", nargs=-1)
@click.pass_context
def configure(ctx, user, bot, extra):
    """
    Configures the .qalx and .bots config files.
    Usage:

    `qalx configure` - configures user and bot default profiles

    `qalx --user-profile=dev configure` - configures dev user profile and
    default bot profile

    `qalx --bot-profile=dev configure --no-user` - only configures bot dev
    profile.  Doesn't configure user profile

    `qalx configure --no-bot customkey=customvalue customkey2=customvalue2`
    - configures user profile with two extra key/values pairs on the config
    """

    if not user and not bot:
        raise ClickException(
            "`user` or `bot` must be specified otherwise no "
            "config files can be created.  Either call "
            "without any arguments or only use a "
            "single `--no-user` or `--no-bot` switch"
        )

    user_profile = ctx.obj["USER_PROFILE"]
    bot_profile = ctx.obj["BOT_PROFILE"]
    check_existing = []

    if user:
        check_existing.append((UserConfig, user_profile, "user"))
    if bot:
        check_existing.append((BotConfig, bot_profile, "bot"))

    for ConfigClass, profile_name, config_type in check_existing:
        # If the profiles already exist we just exit.  It's up to the user
        # to make any changes manually to the profile file - otherwise we risk
        # parsing the file incorrectly and overwriting data the user wants to
        # keep
        try:
            ConfigClass().from_inifile(profile_name)
            config_path = ConfigClass.config_path()
            raise ClickException(
                f"`{profile_name}` profile for config stored "
                f"at `{config_path}` "
                f"already exists.  To make changes you must "
                f"edit the config file directly or specify a "
                f"`--no-{config_type}` flag"
            )
        except (QalxConfigProfileNotFound, QalxConfigFileNotFound):
            # QalxConfigProfileNotFound: The specific profile could not be
            #                            found in the given ConfigClass file
            # QalxConfigFileNotFound: The specific file could not be found
            #                         for the given ConfigClass
            pass

    config_items = {}

    for item in extra:
        # Handles the user being able to pass in extra key/value pairs to write
        # to the config file
        split_item = item.split("=", 1)
        if len(split_item) != 2:
            raise ClickException(
                f"extra arguments must be in the format "
                f"`key=value`.  Got `{item}`"
            )
        config_items[split_item[0].upper()] = split_item[1]

    if user:
        value = click.prompt("qalx Token", type=str)
        config_items["TOKEN"] = value
        UserConfig.configure(user_profile, config_items)
    if bot:
        BotConfig.configure(bot_profile)



import json
import os

import logging
from types import MethodType

from pyqalx.config import UserConfig, BotConfig
from pyqalx.core.entities import QalxListEntity
from pyqalx.core.errors import (
    QalxError,
    QalxNoGUIDError,
    QalxAPIResponseError,
    QalxMultipleEntityReturned,
    QalxEntityNotFound,
    QalxConfigError,
    QalxInvalidSession,
    QalxFileError,
)

logger = logging.getLogger(__name__)


class QalxAdapter(object):
    """
    The base class for a QalxAdapter. An adapter is the interface for the
    entity to the rest api.  This allows us to have a consistent interface
    across all entity types.
    Can be instatiated in two ways:

    QalxAdapter(entity_type='item') -> Returns a :class:`QalxItem` instance

    QalxItem() -> Returns a :class:`QalxItem` instance
    """

    def __init__(self, session, *args, **kwargs):
        QalxSession = __import__("pyqalx").QalxSession
        if not isinstance(session, QalxSession):
            raise QalxInvalidSession(
                f"`qalx_session` should be an instance of"
                f" `{QalxSession}`.  "
                f"Got `{type(session)}`"
            )
        self.session = session
        super(QalxAdapter, self).__init__(*args, **kwargs)

    def __getattribute__(self, item):
        """
        Certain methods can only be called by sessions with a UserConfig or a
        BotConfig.  The API will return a 403 Permission Denied
        but this method handles showing the user a more useful error message
        """
        attr = super(QalxAdapter, self).__getattribute__(item)
        if callable(attr) and isinstance(attr, MethodType):
            # This is a method on `QalxBot`.  Check the config
            # is correct for the method we are calling
            def _msg(expected_class, actual_class):
                return (
                    f"Method `{item}` on `{self.__class__}` "
                    f"must be called via a session with a `{expected_class}`"
                    f"instance.  Got `{actual_class}`"
                )

            is_user_config = isinstance(self.session.config, UserConfig)
            is_bot_config = isinstance(self.session.config, BotConfig)

            if (
                item in getattr(self, "_user_only_methods", [])
                and not is_user_config
            ):
                raise QalxConfigError(_msg(UserConfig, BotConfig))
            if (
                item in getattr(self, "_bot_only_methods", [])
                and not is_bot_config
            ):
                raise QalxConfigError(_msg(BotConfig, UserConfig))
        return attr

    @property
    def entity_class(self):
        return self.session._registry["entities"][
            self._entity_class.entity_type
        ]

    def _process_api_request(self, method, *args, **kwargs):
        """calls to transport layer

        :param method: http method required
        :param args: args to be passed to endpoint method
        :param kwargs: kwargs to be passed to endpoint method
        :returns: `dict` containing API resource data
        """

        headers = kwargs.pop("headers", {})

        try:
            json.dumps(kwargs)
        except (TypeError, ValueError, OverflowError):
            raise QalxError(
                "One of the keyword arguments is not JSON serializable"
            )

        try:
            endpoint = getattr(self.session.rest_api, method.lower())
            logger.debug(str(endpoint))
        except AttributeError:
            raise QalxError(f"{method} not recognised.")

        success, data = endpoint(*args, json=kwargs, headers=headers)

        if success:
            return data
        m = "API request error, message:\n\n-vvv-\n\n"
        m += "\n".join([f"{k}: {v}" for k, v in data.items()])
        m += "\n\n-^^^-"
        raise QalxAPIResponseError(m)

    def detail_endpoint(self, guid, *args, **kwargs):
        """
        The endpoint for interfacing with a single
        instance of `self.entity_class`
        """
        return f"{self.entity_class.entity_type}/{guid}"

    def list_endpoint(self, *args, **kwargs):
        """
        The endpoint for interfacing with multiple instances of
        `self.entity_class`
        """
        return self.entity_class.entity_type

    @staticmethod
    def get_keys_to_save(entity):
        """
        When saving an entity not every key should be saved
        """
        # info & guid are both read only
        # status should only be saved via `update_status` (if available)
        return {
            k: entity[k] for k in entity if k not in ("info", "guid", "status")
        }

    @staticmethod
    def _validate(**kwargs):
        """
        Hook for specifying adapter level validation.  Returns kwargs
        so that if we need to mutate any of the kwargs (i.e. ensuring that
        a key not provided by the user has a specific value) then we can
        """
        if kwargs.get("meta", None) is None:
            # meta is optional but must always be sent through as a dict
            kwargs["meta"] = {}
        if not isinstance(kwargs["meta"], dict):
            raise QalxError("`meta` kwarg must be a `dict`")
        return kwargs

    def add(self, **kwargs):
        """
        Adds a new instance of `self.entity_class`.  Provide valid `kwargs`
        for the object you are trying to create

        :return: An instance of `self.entity_class`
        """
        if kwargs.pop("_is_validated", False) is False:
            # We may have already validated at this point if we are
            # doing a bulk insert.  This kwarg should be set on the adapter
            # that has done the validation
            kwargs = self._validate(**kwargs)
        if "tags" in kwargs.keys():
            # Ensures users don't get confused if they try to do
            # `.add(tags={'my': 'tag'}).  We cannot do this in `_validate`
            # because we only want this validation to occur on `add`
            raise QalxError(
                "Tags can only be added to entities via the "
                "session.  Do `session.tags.add(name=<tag_name>, "
                "value=<tag value>)` rather than passing a `tags` "
                "argument to `add`"
            )
        kwargs["tags"] = self.session.tags
        response = self._process_api_request(
            "post", self.list_endpoint(**kwargs), **kwargs
        )
        # TODO: Should we return packable entiies in an unpacked state?
        return self.entity_class(response)

    def get(self, guid, *args, **kwargs):
        """
        Gets an instance of `self.entity_class` by the given guid.

        :param guid: The guid of the entity to get
        :return: An instance of `self.entity_class`
        """
        endpoint = self.detail_endpoint(guid=guid, *args, **kwargs)
        resp = self._process_api_request("get", endpoint, *args, **kwargs)
        logger.debug(
            f"get {self.entity_class.entity_type} with "
            "guid {guid} with {endpoint}"
        )
        return self.entity_class(resp)

    def find_one(self, query, **kwargs):
        """
        Method for returning a unique entity. Will return the entity that
        matches the query

        :param query: The mongoDB query
        :type query: dict
        :return: an instance of `self.entity_class`
        :raises: `QalxMultipleEntityReturned`
        :raises: `QalxEntityNotFound`
        """
        return self.find(many=False, limit=2, query=query, **kwargs)

    def find(
        self, query=None, sort=None, skip=0, limit=25, many=True, **kwargs
    ):
        """
        Method for listing entities.

        :param query: The optioanl Mongo query to find entities
        :param sort: The keys to sort by
        :param skip: The number of results to skip (offset) by
        :param limit: How many results should the response be limited to
        :param many: Should many entities be returned or just a single one
        :param kwargs: kwargs to query by
        :return:
        """
        list_endpoint = self.list_endpoint(**kwargs)

        logger.debug(
            f"find {self.entity_class.entity_type} with {list_endpoint}"
        )

        if query:
            # We need to dump the query to JSON as there could be
            # complex lookups which would get lost when the request is made
            try:
                query = json.dumps(query)
            except (TypeError, ValueError, OverflowError):
                raise QalxError(
                    "One of the keyword arguments is not JSON serializable"
                )

        resp = self._process_api_request(
            "get",
            list_endpoint,
            sort=sort,
            skip=skip,
            limit=limit,
            query=query,
            **kwargs,
        )
        entities = QalxListEntity(resp, child=self.entity_class)

        if many is False:

            # We are expecting only a single entity.
            entities = entities["data"]
            if len(entities) > 1:
                entities_str = "\n".join([str(q) for q in entities])
                raise QalxMultipleEntityReturned(
                    "Expected one but got "
                    "{}:\n{}".format(len(entities), entities_str)
                )
            elif entities:
                return entities[0]
            else:
                raise QalxEntityNotFound(
                    self.entity_class.entity_type + " not found."
                )
        return entities

    def reload(self, entity, *args, **kwargs):
        """
        Reloads the current entity from the API

        :return: A refreshed instance of `self.entity`
        """
        return self.get(entity["guid"], *args, **kwargs)

    def save(self, entity, *args, **kwargs):
        """
        Saves `entity` to the database.

        :return: The updated instance of `entity`
        """
        if not entity.get("guid"):
            raise QalxNoGUIDError("No guid.")
        # kwargs is also passed through as it could contain data required
        # for validation by any subclasses (blueprints/files etc)
        entity = self._validate(**entity, **kwargs)
        guid = entity["guid"]

        # If there are tags on the session then use them. Otherwise, fallback
        # to using the tags on the entity. This prevents a session without
        # tags from accidentally removing tags previously set on an entity.
        # If there are no tags on the entity (because they weren't returned via
        # `fields` then we don't update the tags - unless tags have been
        # specified on the session)
        tags_to_save = self.session.tags or entity.get("tags", None)
        if tags_to_save:
            entity["tags"] = tags_to_save

        endpoint = self.detail_endpoint(guid=guid, *args, **kwargs)
        keys_to_save = self.get_keys_to_save(entity=entity)
        logger.debug(
            f"save {self.entity_class.entity_type} with "
            "guid {guid} with {endpoint}"
        )

        resp = self._process_api_request("patch", endpoint, **keys_to_save)
        return self.entity_class(resp)

    def archive(self, entity, *args, **kwargs):
        """
        Archives `entity`

        :return: The archived instance of `entity`
        """
        guid = entity["guid"]
        detail_endpoint = self.detail_endpoint(guid=guid, *args, **kwargs)
        endpoint = f"{detail_endpoint}/archive"
        logger.debug(
            f"archive {self.entity_class.entity_type} with "
            "guid {guid} with {endpoint}"
        )

        resp = self._process_api_request("patch", endpoint)
        return self.entity_class(resp)


class QalxUnpackableAdapter(QalxAdapter):
    """
    A qalx adapter for unpackable entities (Set, Group)
    """

    @property
    def kids(self):
        """
        The key that the child entities live in
        """
        return self.child_entity_class.entity_type + "s"

    @staticmethod
    def _child_list_response(child_adapter, query, **kwargs):
        return child_adapter.find(query=query)

    def _child_list_request(
        self, entity, kids, child_adapter, query_key="guid"
    ):
        """
        Builds the list request for getting any unpackable children. Handles
        chunking of the query and also pagination of the response
        :param entity: The entity that has the children
        :param kids: The details of the kids entities we want to lookup
        :param child_adapter: The child adapter to use
        :param query_key: The query key (defaults to `guid`)
        :return: A list of unpacked child_adapter.entity_class entities
        """
        unpacked_entities = []
        for _page in entity._chunks(kids, chunk_size=100):
            # Chunk to avoid hitting maximum request size
            values = list(filter(None, _page))
            query = {query_key: {"$in": values}}
            _resp = self._child_list_response(
                child_adapter=child_adapter, query=query, entity=entity
            )
            unpacked_entities += _resp["data"]
            while _resp["query"]["next"] is not None:
                # We got a paginated response so keep going through the pages
                # until we exhaust this chunk of data
                logger.debug(
                    f"get {self.entity_class.entity_type} with "
                    "{_resp['query']['next']}"
                )
                _resp = self._process_api_request("get", _resp["query"]["next"])
                unpacked_entities += QalxListEntity(
                    _resp, child=self.child_entity_class
                )["data"]
        return unpacked_entities

    def _kids_for_lookup(self, entity):
        """
        For the given entity, will get the child keys so that an API request
        can be made to obtain them
        :param entity: The entity that has child
        :return: The child keys to send to the API, as a list of values
        """
        return list(entity[self.kids].values())

    def _get_child_entities(self, entity):
        """
        Gets all the child entities for the given entity.  Handles chunking
        the children up into blocks to guard against sending a request that is
        too large.  Also handles the paginated result coming back from the
        `list` endpoint.  Does a `list` lookup for self.child_entity_class
        filtering on `guid` of the children to reduce the amount of queries

        :param entity: The entity to be unpacked
        :return: The child entities as a list
        """
        kids = self._kids_for_lookup(entity)

        child_adapter = getattr(
            self.session, self.child_entity_class.entity_type
        )
        unpacked_entities = self._child_list_request(
            entity=entity, kids=kids, child_adapter=child_adapter
        )
        return unpacked_entities

    def _unpacked_entities_to_valid_children(self, entity, unpacked_entities):
        """
        Converts a list of `unpacked_entities` into valid child entities
        for the given entity.

        :param entity: The entity that has kids that should be unpacked
        :param unpacked_entities: The unpacked child entities that need to be
                                  stitched into the entity
        :return: The unpacked entities in the correct format to be assigne to
                 `entity[self.kids]`
        """
        to_return = {}
        for key, item_guid in entity[self.kids].items():
            for child_entity in unpacked_entities:
                if item_guid == child_entity["guid"]:
                    child_entity_adapter = getattr(
                        self.session, child_entity.entity_type
                    )
                    if hasattr(child_entity_adapter, "child_entity_class"):
                        # The child entity has a child adapter of it's own.
                        # Attempt unpack. This can occur if we are unpacking
                        # a Group as we may need to unpack the Sets on the
                        # Group
                        to_return[key] = child_entity_adapter._attempt_unpack(
                            child_entity
                        )
                    else:
                        to_return[key] = child_entity
        return to_return

    def _should_unpack(self):
        return self.session.config.getboolean(
            "UNPACK_" + self.entity_class.entity_type.upper()
        )

    def _attempt_unpack(self, entity):
        """
        For the given entity will attempt
        to unpack the child_entity data that it contains

        :entity: The entity to be unpack
        :return: The entity instance with the child_entity objects unpacked
        """

        if self._should_unpack() and self.kids in entity:
            # `self.kid` might not be there if we have specified a subset
            # of fields
            unpacked_entities = self._get_child_entities(entity)
            entity[self.kids] = self._unpacked_entities_to_valid_children(
                entity=entity, unpacked_entities=unpacked_entities
            )
        return entity

    def _pack_kids_for_request(self, kwargs):
        """
        Packs the kids into a format that they need to be in
        for sending to the API
        :param kwargs: The kwargs sent to the `add` method
        :return: kwargs with packed kids
        """
        kids = {
            key: str(kid["guid"]) for key, kid in kwargs[self.kids].items()
        }  # noqa
        kwargs[self.kids] = kids
        return kwargs

    def get_keys_to_save(self, entity):
        """
        Unpackable entities will have unpacked data stored on them that have
        to be submitted as packed data
        """
        keys_to_save = super(QalxUnpackableAdapter, self).get_keys_to_save(
            entity=entity
        )
        if self.kids in keys_to_save.keys():
            keys_to_save[self.kids] = self.get_kids_keys_to_save(
                keys_to_save[self.kids]
            )
        return keys_to_save

    @staticmethod
    def get_kids_keys_to_save(kids_to_save):
        """
        Gets the keys of the kids that we want to save.  Ensures that if the
        keys are in an packed state we don't attempt to unpack them.
        :param kids_to_save: A dict of the kids we want to save
        :type kids_to_save: dict
        :return: A dict of {key:guid} of the kids we want to save
        """
        return {
            k: str(i["guid"]) if isinstance(i, dict) else i
            for k, i in kids_to_save.items()
        }

    @staticmethod
    def cache_original_kids(kids):
        """
        When saving the API only wants to know about the guids of the kids so
        we need to cache the originals
        until after the save
        :param kids: The kids on this entity
        :type kids: dict
        :return: A dictionary of the kids
        """
        return {k: v for k, v in kids.items()}

    def get(self, guid, *args, **kwargs):
        """
        Unpackable entities need to be unpacked after retrieval

        :param guid: The `guid` of the entity to get
        :return: An unpacked entity
        """
        entity = super(QalxUnpackableAdapter, self).get(
            guid=guid, *args, **kwargs
        )
        entity = self._attempt_unpack(entity)
        return entity

    def find(self, many=True, *args, **kwargs):
        entities = super(QalxUnpackableAdapter, self).find(
            many=many, *args, **kwargs
        )
        if many is False:
            # Only unpack a list view if we have a single entity
            entities = self._attempt_unpack(entities)
        return entities

    def save(self, entity, *args, **kwargs):
        """
        When saving an unpacked entity the kids need to be packed.  To save
        having to unpack them again just save the original kids and replace
        them with the packed values after saving
        """
        has_kids_key = self.kids in entity.keys()
        if has_kids_key:
            original_kids = self.cache_original_kids(entity[self.kids])

        entity = super(QalxUnpackableAdapter, self).save(
            entity=entity, *args, **kwargs
        )
        if has_kids_key:
            entity[self.kids] = original_kids
        else:
            entity = self._attempt_unpack(entity=entity)
        return entity

    def add(self, **kwargs):
        # Need to validate before we pack the kids up
        kwargs = self._validate(**kwargs)
        kwargs = self._pack_kids_for_request(kwargs)
        return super(QalxUnpackableAdapter, self).add(
            _is_validated=True, **kwargs
        )


class QalxSignalAdapter(QalxAdapter):
    """
    Bots and Workers have signals that are used to determine when to stop
    processing data.  This class provides that shared functionality
    """

    def get_signal(self, entity, *args, **kwargs):
        """
        Gets just the `signal` field from the entity and then parses that
        into the `signal_class`
        """
        entity = self.get(
            guid=entity["guid"], fields="signal", *args, **kwargs
        )  # noqa
        signal = self.signal_class(entity)
        return signal

    def terminate(self, entity, *args, **kwargs):
        """
        Updates the entity with a terminate signal
        """
        guid = entity["guid"]
        endpoint = self.detail_endpoint(guid=guid, **kwargs)
        logger.debug(
            f"terminate {self.entity_class.entity_type} with guid {guid} with {endpoint}"  # noqa
        )
        signal = self.signal_class._terminate_signal(*args, **kwargs)

        self._process_api_request("patch", endpoint, signal=signal)

    def _stop_or_resume(self, entity, stop, **kwargs):
        guid = entity["guid"]
        endpoint = self.detail_endpoint(guid=guid, **kwargs)
        signal = self.signal_class._stop_signal(stop)
        logger.debug(
            f"signal {signal} {self.entity_class.entity_type} with guid {guid} with {endpoint}"  # noqa
        )
        self._process_api_request("patch", endpoint, signal=signal)

    def stop(self, entity, **kwargs):
        """
        Updates the entity with a stop signal
        """
        self._stop_or_resume(entity, stop=True, **kwargs)

    def resume(self, entity, **kwargs):
        """
        Updates the entity with a resume signal
        """
        self._stop_or_resume(entity, stop=False, **kwargs)


class QalxNamedEntityAdapter(QalxAdapter):
    """
    Certain QalxEntities have a `name` field that can be used to look up the
    entity.
    """

    def get_by_name(self, name):
        """a single entity by name

        :param name: name of entity
        :type name: str
        :return: instance of `self.entity_class`
        :raises: pyqalx.errors.QalxReturnedMultipleError,
                 pyqalx.errors.QalxEntityNotFound
        """
        return self.find_one(query={"name": name})

    def get_or_create(self, name, meta=None, **kwargs):
        """
        Gets an Entity by the given name or creates it if it doesn't exist

        :param name: The `name` that to use to get or create the entity by
        :type name: str
        :param meta: metadata about the entity
        :param kwargs: Other kwargs on the entity to use when creating
        :return: instance of `self.entity_class`
        """
        try:
            return self.get_by_name(name=name)
        except QalxEntityNotFound:
            return self.add(name=name, meta=meta, **kwargs)


class QalxValidateBlueprintAdapter(QalxAdapter):
    """
    Certain entity types can be validated against a preexisting Blueprint
    when they are created.
    """

    def _validate(self, blueprint_name=None, **kwargs):
        """
        When adding entities that inherit from this class, pyqalx will get
        the blueprint based on the `blueprint_name` and then ensure that
        the schema validates correctly against the data that you are creating
        the entity with

        :param blueprint_name: the name of the blueprint
        :type blueprint_name: str, None
        :param get_blueprint: Should this method retrieve the blueprint?
        :type get_blueprint: bool, False
        :return: A valid `self.entity_class` instance
        :raises: jsonschema.ValidationError
        """
        validated_data = super(QalxValidateBlueprintAdapter, self)._validate(
            **kwargs
        )
        if blueprint_name is not None:
            blueprint = validated_data.pop("blueprint", None)
            if blueprint is None:
                blueprint = self._get_blueprint(name=blueprint_name)
            blueprint.validate(validated_data)
        return validated_data

    def _get_blueprint(self, name):
        """
        Helper method for getting the blueprint instance
        :param name: The name of the blueprint
        :type name: str
        :return: A `~entities.blueprint.Blueprint` instance
        """
        blueprint = self.session.blueprint.get_by_name(name=name)
        return blueprint

    def add(self, blueprint_name, **kwargs):
        """
        When adding entities that inherit from this class, pyqalx will get
        the blueprint based on the `blueprint_name` and then ensure that
        the schema validates correctly against the data that you are creating
        the entity with

        :param blueprint_name: the name of the blueprint
        :type blueprint_name: str, None
        :return: A valid `self.entity_class` instance
        :raises: jsonschema.ValidationError
        """
        return super(QalxValidateBlueprintAdapter, self).add(
            blueprint_name=blueprint_name, **kwargs
        )

    def save(self, entity, blueprint_name=None, *args, **kwargs):
        """
        When saving entities that inherit from this class, pyqalx will get
        the blueprint based on the `blueprint_name` and then ensure that
        the schema validates correctly against the data that you are updating
        the entity with
        :param entity: The entity instance that you are saving
        :type entity: entities.QalxEntity instance
        :param blueprint_name: the name of the blueprint
        :type blueprint_name: str, None
        :return: A valid `self.entity_class` instance
        :raises: jsonschema.ValidationError
        """
        return super(QalxValidateBlueprintAdapter, self).save(
            entity, blueprint_name=blueprint_name, **kwargs
        )


class QalxFileAdapter(QalxAdapter):
    """
    Adapter that allows files to be uploaded to an entity
    """

    def _validate(self, input_file, file_name, **kwargs):
        """
        Ensures that the `input_file` and/or `file_name` are present
        and correct
        :param input_file: The path to the file or the file stream
        :param file_name: The name of the file
        :return: kwargs
        """
        if input_file is not None:
            # Only validate the file is it's specified in the
            # implementing adapter
            is_filestream = self.session.rest_api.is_filestream(input_file)
            if is_filestream:
                if not file_name:
                    raise QalxFileError(
                        "You must supply a file name when supplying"
                        " a file stream"
                    )
                file = {"name": file_name}
            else:
                if os.path.isfile(input_file):
                    file_path = os.path.abspath(input_file)
                    filename = (
                        file_name if file_name else os.path.basename(file_path)
                    )
                    file = {"name": filename}
                else:
                    raise QalxFileError(
                        "Invalid file `{0}`".format(input_file)
                    )  # noqa
            kwargs["file"] = file
        return super(QalxFileAdapter, self)._validate(**kwargs)

    def add(self, input_file, file_name, upload=True, **kwargs):
        """
        When a user creates an entity with a file pyqalx will upload the file
        automatically for you after the entity itself has been created.  Pass
        `upload=False` through to stop this from happening and upload the file
        yourself
        :param input_file: The input file path or stream
        :param file_name: The optional file name
        :param upload: Whether the file should be automatically uploaded or not
        :return: A valid entity
        """
        entity = super(QalxFileAdapter, self).add(
            input_file=input_file, file_name=file_name, **kwargs
        )
        if upload and input_file is not None:
            entity = self._upload_file(entity=entity, input_file=input_file)
        return entity

    def save(
        self, entity, input_file, file_name, upload=True, *args, **kwargs
    ):  # noqa
        """
        Saves an entity - optionally updating any file data.
        :param entity: The entity that is being saved
        :param input_file: The input file path or stream
        :param file_name: The optional file name
        :param upload: Whether the file should be automatically uploaded or not
        :return: A valid entity
        """
        entity = super(QalxFileAdapter, self).save(
            entity=entity, input_file=input_file, file_name=file_name, **kwargs
        )
        if input_file is not None:
            if upload:
                entity = self._upload_file(
                    entity=entity, input_file=input_file
                )  # noqa
            else:
                # Entity is returned with the put_url so that the user can
                # upload the file manually
                return entity
        # Ensure "put_url" is removed even if we are not changing the file
        entity["file"].pop("put_url", None)
        return entity

    def _upload_file(self, entity, input_file):
        # Only upload if everything was OK with the item creation
        # and the user wants to upload the file automatically.
        s3_resp_ok, s3_data = self.session.rest_api._upload_to_s3(
            entity, input_file
        )  # noqa
        if s3_resp_ok is False:
            # The S3 upload failed, return the failed S3 response in place
            # of the response from the api.  Any implementing client will
            # have to handle tidying up any incorrect data.
            # TODO: Handle errors in a consistent way in adapters
            m = "API request error, message:\n\n-vvv-\n\n"
            m += "\n".join([f"{k}: {v}" for k, v in s3_data.items()])
            m += "\n\n-^^^-"
            raise QalxAPIResponseError(m)
        else:
            # Always pop the `put_url` off if it's a success
            entity["file"].pop("put_url")
        return entity


class QalxUpdateStatusAdapter(QalxAdapter):
    def update_status(self, entity, status, **kwargs):
        """
        Updates the status of an entity

        :param entity: The entity that is being updated
        :type entity: An instance of ~entities.entity.QalxEntity
        :param status: The status to update to
        :type: status: str
        """
        guid = entity["guid"]
        endpoint = self.detail_endpoint(guid=guid, **kwargs)
        logger.debug(
            f"update status{self.entity_class.entity_type} with guid {guid} with {endpoint}"  # noqa
        )
        status_dict = {"message": status}
        self._process_api_request("patch", endpoint, status=status_dict)



from .core import BasePyQalxAPI, PyQalxAPIException


class PyQalxAPI(BasePyQalxAPI):
    def get(self, endpoint, **kwargs):
        """
        Performs a GET request to the specified endpoint
        :param endpoint: The endpoint you want to query: 'item/<item_guid>'
        :param kwargs: Any arguments you want to pass to the request:
                json={'data': 'filter__by=gte=4, 'limit':5, 'sort':<sort>, }
        :return: response
        """
        url = self._get_url(endpoint)
        if kwargs.get("json", None) and kwargs.get("params", None):
            # They have supplied params and json.  Raise an exception as we
            # can't safely merge them
            raise PyQalxAPIException(
                "qalx GET parameters must be supplied"
                "using the `json` parameter for API"
                "consistency"
            )
        json_removed = kwargs.pop("json", None)
        if json_removed and kwargs.get("params") is None:
            # pyqalx will send `params` via the `json` argument for API
            # consistency.  Any user using the transport layer on its own
            # can send either `json` or `params`.  The querystring cannot
            # be in the request body
            kwargs["params"] = json_removed
        return self._build_request(url=url, method="GET", **kwargs)

    def post(self, endpoint, **kwargs):
        """
        Performs a POST request to the specified endpoint
        :param endpoint: The endpoint you want to query: 'item'
        :param kwargs: Any data you want to pass to the request.
        :return: response
        """
        return self._build_non_safe_request(
            method="POST", endpoint=endpoint, **kwargs
        )

    def patch(self, endpoint, **kwargs):
        """
        Performs a PATCH request to the specified endpoint
        :param endpoint: The endpoint you want to query: 'item/<item_guid>'
        :param kwargs: Any data you want to pass to the request.
        :return: response
        """
        return self._build_non_safe_request(
            method="PATCH", endpoint=endpoint, **kwargs
        )

    def put(self, endpoint, **kwargs):
        """
        Performs a PUT request to the specified endpoint
        :param endpoint: The endpoint you want to query: 'item/<item_guid>'
        :param kwargs: Any data you want to pass to the request.
        :return: response
        """
        return self._build_non_safe_request(
            method="PUT", endpoint=endpoint, **kwargs
        )

    def delete(self, endpoint):
        """
        Performs a DELETE request to the specified endpoint
        :param endpoint: The endpoint you want to query: 'item/<item_guid>'
        """
        url = self._get_url(endpoint)
        return self._build_request(url=url, method="DELETE")



import platform

from logging import getLogger

from pyqalx.config import UserConfig
from pyqalx.core.adapters import (
    QalxBlueprint,
    QalxBot,
    QalxGroup,
    QalxItem,
    QalxQueue,
    QalxSet,
    QalxWorker,
)
from pyqalx.core.log import configure_logging
from pyqalx.core.registry import Registry
from pyqalx.core.tags import Tags
from pyqalx.transport import PyQalxAPI


class QalxSession:
    """The session that any interaction with the API will use."""

    def __init__(
        self,
        profile_name="default",
        config_class=None,
        skip_ini=False,
        rest_api_class=None,
        project_name=None,
    ):
        """
        :param profile_name: profile name to get from `config_class` (default="default")
        :type profile_name: str
        :config_class: The class for the config that will be used for this session
        :param skip_ini: Should loading the config from the inifile be skipped
        :type skip_ini: bool
        :param rest_api_class: The class to use for the rest api
        :param tags: A list of dicts for tags that you want to add to every
                     entity that is added/saved during this session
        :type tags: list
        """

        if config_class is None:
            config_class = UserConfig
        self.config = config_class()
        """an instance of the `config_class` (default=pyqalx.config.UserConfig())"""
        if not skip_ini:
            self.config.from_inifile(profile_name)

        configure_logging(self.config)
        if rest_api_class is None:
            rest_api_class = PyQalxAPI
        self.rest_api = rest_api_class(self.config)
        self._registry = Registry()
        self._registry._register_defaults()
        self.tags = Tags([], rest_api=self.rest_api)
        if project_name:
            self.tags.add(name="project_name", value=project_name)

    def __setattr__(self, key, value):
        if key == "tags" and hasattr(self, "tags"):
            # The user is completely replacing the tags.
            # Attempt to validate them before replacing them.
            self.tags._validate(value)
        return super(QalxSession, self).__setattr__(key, value)

    @property
    def log(self):
        return getLogger("pyqalx.integration")

    def _update_config(self, config):
        """
        Method to use if the config needs to be updated after the session has
        been created.  Used when creating a `~bot.Bot` as a `QalxSession`
        needs to be created using a `BotConfig` and then update it with the
        token the `~bot.Bot` needs to use to interact with the
        `~entities.Queue`.
        Also updates the rest_api with the updated config

        :param config:
        :return:
        """
        self.config.update(config)
        self.rest_api = self.rest_api.__class__(self.config)

    def register(self, cls):
        self._registry.register(cls)

    def unregister(self, cls):
        self._registry.unregister(cls)

    @property
    def _host_info(self):
        return {
            "node": platform.node(),
            "platform": platform.platform(),
            # TODO: add more platform and IP address infos
        }

    @property
    def item(self):
        """
        returns a :class:`~pyqalx.core.adapters.item.QalxItem` adapter for this session

        :return: pyqalx.core.adapters.item.QalxItem
        """
        return QalxItem(self)

    @property
    def set(self):
        """
        returns a :class:`~pyqalx.core.adapters.set.QalxSet` adapter for this session

        :return: pyqalx.core.adapters.set.QalxSet
        """
        return QalxSet(self)

    @property
    def group(self):
        """
        returns a :class:`~pyqalx.core.adapters.group.QalxGroup` adapter for this session

        :return: pyqalx.core.adapters.group.QalxGroup
        """
        return QalxGroup(self)

    @property
    def queue(self):
        """
        returns a :class:`~pyqalx.core.adapters.queue.QalxQueue` adapter for this session

        :return: pyqalx.core.adapters.queue.QalxQueue
        """
        return QalxQueue(self)

    @property
    def bot(self):
        """
        returns a :class:`~pyqalx.core.adapters.bot.QalxBot` adapter for this session

        :return: pyqalx.core.adapters.bot.QalxBot
        """
        return QalxBot(self)

    @property
    def worker(self):
        """
        returns a :class:`~pyqalx.core.adapters.worker.QalxWorker` adapter for this session

        :return: pyqalx.core.adapters.worker.QalxWorker
        """
        return QalxWorker(self)

    @property
    def blueprint(self):
        """
        returns a :class:`~pyqalx.core.adapters.blueprint.QalxBlueprint` adapter for this session

        :return: pyqalx.core.adapters.blueprint.QalxBlueprint
        """
        return QalxBlueprint(self)



from pyqalx.core.adapters.adapter import (
    QalxUnpackableAdapter,
    QalxValidateBlueprintAdapter,
)
from pyqalx.core.entities import Item, Set


class QalxSet(
    QalxValidateBlueprintAdapter, QalxUnpackableAdapter,
):
    _entity_class = Set
    child_entity_class = Item

    def add(self, items, meta=None, blueprint_name=None, **kwargs):
        """
        When adding a `Set` ensure that the items posted to the api are in the
        format {<key>: <Item instance>}

        :param items: A dictionary of Items to create on the set
        :type items: dict
        :param meta: A dictionary of metadata to store on this set
        :type meta: dict
        :param blueprint_name: An optional blueprint
         name to use if you want to validate this set against an existing Blueprint
        :type blueprint_name: str
        :return: A newly created `Set` instance

        """
        return super(QalxSet, self).add(
            items=items, blueprint_name=blueprint_name, meta=meta, **kwargs
        )



import json
import threading
from datetime import datetime, timezone
import uuid

import boto3
import dateutil
import logging
from botocore.exceptions import ClientError

from pyqalx.core.entities import QalxEntity, Group, Item, Set
from pyqalx.core.errors import QalxQueueError


logger = logging.getLogger(__name__)


class QueueResponse:
    """a response from a remote queue"""

    def __init__(self, response):
        """make new response

        v0.1 - doesn't do much yet
        :param response: response from queue
        """
        self._raw_response = response


class QueueMessage:
    """a message frm the queue

    attributes:
        body: message body
        receipt_handle: some information to return to the message broker to
        confirm receipt
    """

    def __init__(self, message):
        """

        :param message: the full raw message from the queue
        """
        self._raw_message = message
        self.body = json.loads(message["Body"])
        self.receipt_handle = message["ReceiptHandle"]
        self._heartbeats = 1
        self._heartbeat = None

    def _stop_heartbeat(self):
        if self._heartbeat is not None:
            self._heartbeat.set()
            # Make sure that if we start the heartbeat again on this message
            # it starts from 1
            self._heartbeats = 1

    def _do_heartbeat(self, visibility, queue):
        """
        A QalxJob might take a long time to process.  Therefore, we
        use a heartbeat on the message to keep it in flight for as long
        as possible.
        :param visibility: The `MSG_BLACKOUTSECONDS` from the bot_config.
        :param queue: An instance of :class:~queue.Queue
        :return:
        """
        if self._heartbeat is None or self._heartbeat.is_set() is False:
            # We only do the heartbeat if it hasn't been started yet and we
            # haven't unset it
            self._heartbeat = threading.Event()

            # Ensure it's an int - it might not be if it's come from the config
            visibility = int(visibility)

            visibility = visibility * self._heartbeats

            # The max timeout for a message is 12 hours in seconds
            MAX_VISIBILITY_TIMEOUT = 43200
            if visibility <= MAX_VISIBILITY_TIMEOUT:
                # Don't change visibility the first time as the message will
                # be in flight anyway
                try:
                    queue.broker_client.change_message_visibility(
                        QueueUrl=queue["info"]["queue_url"],
                        ReceiptHandle=self.receipt_handle,
                        VisibilityTimeout=visibility,
                    )
                except ClientError as exc:
                    # ClientError: There was an issue changing the visibility.
                    #    This can happen if we delete the message at the same
                    #    time as we are changing the visibility (which is
                    #    possible as the heartbeat is a thread within the
                    #    Worker process).  Just return and let the Worker
                    #    handle the heartbeat again (probably resulting in
                    #    stopping it)
                    logger.debug(
                        f"Got ClientError when extending heartbeat.  "
                        f"Error was: `{exc}`"
                    )
                    # Wait for a single second to help prevent multiple
                    # requests being fired to AWS
                    self._heartbeat.wait(1)
                    return

            # We keep track of the number of heartbeats this message has had
            # and simply increase the visibility timeout exponentially
            self._heartbeats += 1
            # Get roughly 90 percent of the updated visibility time.
            # Sleep until that time expires and then increase the
            # message visibility the next time this method is called
            ninety_percent = visibility - int(visibility / 10)
            self._heartbeat.wait(ninety_percent)


class Queue(QalxEntity):
    """QalxEntity with entity_type Queue

    attributes:
        broker_client: an authenticated client to communicate with the remote
        message broker
    """

    entity_type = "queue"

    @staticmethod
    def _stringify(message: dict) -> str:
        return json.dumps(message)

    def __init__(self, *args, **kwargs):
        super(Queue, self).__init__(*args, **kwargs)
        self.__super_setbroker_client__()

    def __super_setbroker_client__(self, broker_client=None):
        # Use __super_setattr__ to ensure that the `_broker_client` is
        # set as an attribute - not an ObjectDict
        self.__super_setattr__("_broker_client", broker_client)

    @property
    def broker_client(self):
        """
        Windows can't pickle the client if it gets set on __init__ therefore,
        we set it to an internal attribute and cache it for the duration of
        the Queue to avoid getting the client multiple times
        """
        if (
            self._broker_client is None
            and self.get("info")
            and ("credentials" in self["info"])
        ):

            _broker_client = boto3.client(
                "sqs",
                region_name="eu-west-2",  # TODO: #2756 make this configurable
                aws_access_key_id=self["info"]["credentials"]["ACCESS_KEY_ID"],
                aws_secret_access_key=self["info"]["credentials"][
                    "SECRET_ACCESS_KEY"
                ],
                aws_session_token=self["info"]["credentials"]["SESSION_TOKEN"],
            )
            # Use __super_setattr__ to ensure that the `_broker_client` is
            # set as an attribute - not an ObjectDict
            self.__super_setbroker_client__(_broker_client)
        return self._broker_client

    def _credentials_expiring_soon(self):
        """
        Returns True if the credentials are expiring soon.  The expiration
        datetime will be a UTC timestamp in isoformat
        """
        expiration = dateutil.parser.parse(
            self["info"]["credentials"]["EXPIRATION"]
        )
        # Refresh when 1 hour or less left on the expiration token
        REFRESH_LIMIT_HOURS = 3600
        return (
            expiration - datetime.now(timezone.utc)
        ).seconds <= REFRESH_LIMIT_HOURS

    def submit_sets_from_group(self, group):
        """

        :param group: a group
        :type group: `~pyqalx.core.Group`
        :returns: `pyqalx.core.queue.QueueResponse`
        """

        responses = []

        for chunk in self._chunks(list(group["sets"].values()), chunk_size=10):
            chunked_objects = filter(None, chunk)
            entries = []
            for _n, set_guid in enumerate(chunked_objects):
                entries.append(
                    {
                        "Id": str(uuid.uuid4()),
                        "MessageBody": json.dumps(
                            {
                                "entity_type": "set",
                                "entity_guid": set_guid,
                                "parent_group_guid": group["guid"],
                            }
                        ),
                        # Unique MessageGroupId to prevent processing
                        # duplicates
                        "MessageGroupId": str(uuid.uuid4()),
                        # Unique MessageDeduplicationId to allow the user
                        # to put the same entity on the queue multiple times
                        "MessageDeduplicationId": str(uuid.uuid4()),
                    }
                )

            if entries:
                responses.append(
                    self.broker_client.send_message_batch(
                        QueueUrl=self["info"]["queue_url"], Entries=entries
                    )
                )

        return QueueResponse(responses)

    def get_messages(self, max_num_msg, visibility, waittime):
        """
        get messages from the queue

        :param max_num_msg: maximum number to retrieve
        :param visibility: time in seconds until the message becomes visible
                           again on the queue
        :param waittime: time to wait for a message
        :return: list of `pyqalx.core.entities.queue.QueueMessage`
        """

        responses = self.broker_client.receive_message(
            QueueUrl=self["info"]["queue_url"],
            MaxNumberOfMessages=max_num_msg,
            VisibilityTimeout=visibility,
            WaitTimeSeconds=waittime,
        )

        if responses.get("Messages"):
            return [
                QueueMessage(response) for response in responses["Messages"]
            ]
        return []

    def delete_message(self, message):
        """remove message from the queue

        :param message: message to delete
        :type message: `QalxMessage`
        :return: None
        """
        if message is not None:
            # message could be None if the job it's been
            # on has been told to wait - therefore, don't attempt delete
            self.broker_client.delete_message(
                QueueUrl=self["info"]["queue_url"],
                ReceiptHandle=message.receipt_handle,
            )

    def submit_entity(self, entity):
        """put a qalx entity on the queue

        :param entity: entity to add to the queue
        :type entity: `pyqalx.core.entity.QalxEntity`
        :return: `pyqalx.core.queue.QueueResponse`
        """
        if not isinstance(entity, (Item, Set, Group)):
            raise QalxQueueError(
                "Only Item, Set and Group entities can be submitted to the "
                "queue."
            )
        response = self.broker_client.send_message(
            QueueUrl=self["info"]["queue_url"],
            # Unique MessageGroupId to prevent processing duplicates
            MessageGroupId=str(uuid.uuid4()),
            # Unique MessageDeduplicationId to allow the user
            # to put the same entity on the queue multiple times
            MessageDeduplicationId=str(uuid.uuid4()),
            MessageBody=self._stringify(
                {
                    "entity_type": entity.entity_type,
                    "entity_guid": entity["guid"],
                }
            ),
        )
        return QueueResponse(response)
